/**
 * Generate a sdoc from the YUI documentation (extracted from the sources)
 */
package yuisdoc;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @todo I should preserve links of this form:
 *   <a href="YAHOO.util.Element.html#addListener">
 *  or possibly rewrite them as @see tags
 * 
 * @author Tor Norbye
 */

public class Main {

    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("Usage: Main <yui-source-distro>");
        }
        File yui = new File(args[0]);
        if (!yui.exists()) {
            System.err.println(yui.getPath() + " does not exist");
            System.exit(0);
        }
        
        
        try {
            File build = new File(yui, "build");
            assert build.exists() : build.getPath();
        
            List<YuiFile> files = new ArrayList<YuiFile>();            
            new Main().process(build, files, "");

            File root = new File("build/yui");
            root.mkdirs();
            
            for (YuiFile f : files) {
                String name = f.relativeFile;
                assert name.endsWith(".js");
                name = name.substring(0, name.length()-2) + "sdoc";
                File dest = new File(root, name);
                dest = dest.getAbsoluteFile();
                File parent = dest.getParentFile();
                parent.mkdirs();
                Writer writer = new BufferedWriter(new FileWriter(dest));
                writeHeader(writer);
                
                for (String s : f.annotatedComments) {
                    writer.write(s);
                    writer.write("\n");
                }

                writer.close();
            }
        } catch (IOException ex) {
            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    class YuiFile {
        File file;
        String relativeFile;
        String copyright;
        List<String> comments;
        Map<String,String> nextLines;
        List<String> annotatedComments;
    }
    
    private static void writeHeader(Writer writer) throws IOException {
//        writer.write("// See http://developer.yahoo.com/yui/license.html\n");
//
//        writer.write("/*\n" +
//            "Software License Agreement (BSD License)\n" +
//            "Copyright (c) 2008, Yahoo! Inc.\n" +
//            "All rights reserved.\n" +
//            "\n" +
//            "Redistribution and use of this software in source and binary forms, with or without modification,\n" +
//            "are permitted provided that the following conditions are met:\n" +
//            "\n" +
//            "    * Redistributions of source code must retain the above copyright notice, this list of\n" +
//            "      conditions and the following disclaimer.\n" +
//            "    * Redistributions in binary form must reproduce the above copyright notice, this list \n" +
//            "      of conditions and the following disclaimer in the documentation and/or other materials\n" + 
//            "      provided with the distribution.\n" +
//            "    * Neither the name of Yahoo! Inc. nor the names of its contributors may be used to\n" +
//            "      endorse or promote products derived from this software without specific prior written\n" +
//            "      permission of Yahoo! Inc.\n" +
//            "\n" +
//            "THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES,\n" +
//            "INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE\n" + 
//            "ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, \n" + 
//            "INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF\n" +
//            "SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n" + 
//            "ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)\n" +
//            "ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n*/\n\n");
        
        writer.write("/*\n");
        writer.write("Copyright (c) 2008, Yahoo! Inc. All rights reserved.\n");
        writer.write("Code licensed under the BSD License:\n");
        writer.write("http://developer.yahoo.net/yui/license.txt\n");
        writer.write("version: 2.5.0\n");
        writer.write("*/\n");

        writer.write("/* Doc definitions for YUI support in NetBeans, \n"+
                " * automatically generated from the YUI sources\n" +
                " * using the generator in misc/javascript.generatestubs/yui-sdoc\n" +
                " * Version: 1.0\n" +
                " */\n\n");
    }

    void process(File f, List<YuiFile> files, String relative) throws IOException {
        if (f.isDirectory()) {
            for (File c : f.listFiles()) {
                String childDir = relative.length() > 0 ? relative + "/" + c.getName() : c.getName();
                process(c, files, childDir);
            }
            
            return;
        }
        
        if (!f.getName().endsWith(".js")) {
            return;
        }

        if (f.getName().endsWith("-min.js") || f.getName().endsWith("-debug.js")) {
            String base = f.getName();
            base = base.substring(0, base.lastIndexOf('-')) + ".js";
            File sibling = new File(f.getParentFile(), base);
            assert sibling.exists();
            return;
        }
        
        String js = readFile(f);
        
        List<String> comments = new ArrayList<String>();
        Map<String,String> nextLines = new HashMap<String,String>();
        
        findComments(js, comments, nextLines);

        YuiFile yuiFile = new YuiFile();
        yuiFile.file = f;
        yuiFile.relativeFile = relative;
        yuiFile.comments = comments;
        yuiFile.nextLines = nextLines;
        yuiFile.annotatedComments = annotateComments(comments, nextLines);
        
        files.add(yuiFile);
    }

    private List<String> annotateComments(List<String> comments, Map<String,String> nextLines) {
        List<String> result = new ArrayList<String>();

        String namespace = null;
        String clz = null;
        
        for (String s : comments) {
            String original = s;
            ElementType type = null;
            s = fixTypos(s);
            s = removeLinks(s);
            s = fixTypes(s);
            String v = getProperty("@namespace", s);
            if (v != null) {
                namespace = v;
            }
            v = getProperty("@class", s);
            if (v != null) {
                clz = v;
            }
            String name = getProperty("@property", s);
            if ("0".equals(name) || "1".equals(name)) {
                // Strange properties on YAHOO.util.Region...
                continue;
            }
            if (name == null) {
                name = getProperty("@attribute", s);
                if (name == null) {
                    // XXX Is a @config like a property? Check!
                    name = getProperty("@config", s);
                }
            }
            if (name != null) {
                type = ElementType.PROPERTY;
            } else {
                name = getProperty("@event", s);
                if (name != null) {
                    type = ElementType.EVENT;
                } else {
                    name = getProperty("@method", s);
                    if (name != null) {
                        type = ElementType.FUNCTION;
                    } else {
                        if (getProperty("@constructor", s) != null) {
                            assert clz != null;
                            type = ElementType.CONSTRUCTOR;
                        } else {
                            if (s.indexOf("@class") != -1) {
                                assert clz != null;
                                type = ElementType.CONSTRUCTOR;
                            } else if (s.indexOf("@private") != -1) {
                                // YUI has a bunch of comments on methods that don't seem to correspond to APIs,
                                // ignore these
                                assert s.indexOf("@class") == -1 && s.indexOf("@namespace") == -1;
                                continue;
                            } else if (s.indexOf('\n') == -1 && s.startsWith("/************")) {
                                // Just a dumb separator line...
                                continue;
                            } else if (s.indexOf("* Validates the calendar widget.") != -1) {
                                // Known broken documentation
                                type = ElementType.FUNCTION;
                                name = "validate";
                            } else {
                                String nextLine = nextLines.get(original);
                                if (nextLine != null) {
                                    name = nextLine;
                                    String tag = null;
                                    if (name.startsWith("on")) {
                                        type = ElementType.EVENT;
                                        tag = "@event";
                                    } else if (name.startsWith("_get") || name.startsWith("_set")) {
                                        type = ElementType.FUNCTION;
                                        tag = "@method";
                                    } else if (name.indexOf("Event") != -1 || name.indexOf("event") != -1) {
                                        type = ElementType.EVENT;
                                        tag = "@event";
                                    } else if (name.indexOf("ttribute") != -1 || name.indexOf("roperty") != -1) {
                                        type = ElementType.PROPERTY;
                                        tag = "@property";
                                    } else {
                                        type = ElementType.FUNCTION;
                                        tag = "@method";
                                    }
                                    int insert = s.indexOf("*/");
                                    s = s.substring(0, insert) + "* " + tag + " " + name + "\n " + s.substring(insert);
                                } else {
                                    // Known exceptions
                                    if (s.indexOf("Initializes the custom events for Module which are fired") != -1) {
                                        name = "initEvents";
                                        type = ElementType.FUNCTION;
                                    } else if (s.indexOf("The old Calendar2up class is no longer necessary, since CalendarGroup") != -1) {
                                        clz = "YAHOO.widget.Calendar2up";
                                        type = ElementType.CONSTRUCTOR;
                                    } else if (s.indexOf("If the proxy element has no background-color, then it is considere") != -1) {
                                        // Not a real /** */ document - just implementation comment, skip
                                        continue;
                                    } else if (s.indexOf("Placing the IFRAME element inside seems to stop this issue") != -1) {
                                        continue;
                                    } else if (s.indexOf("Name for the test case.") != -1) {
                                        continue;
                                    } else if (s.indexOf("Special rules for the test case. Possible subobjects") != -1) {
                                        continue;
                                    } else if (s.indexOf("Break selector into token units per simple selector.") != -1) {
                                        continue;
                                    } else if (s.indexOf(" CUSTOM EVENTS") != -1) {
                                        continue;
                                    } else if (s.indexOf("Creates a rich Toolbar widget based on Button") != -1) {
                                        // This is a module definition -- for now, skip
                                        continue;
                                    } else if (s.indexOf("@beta") != -1 || s.indexOf("@browser") != -1) {
                                        // Browser and beta workaround comments
                                        continue;
                                    } else {
                                        assert getProperty("@module", s) != null : "No known name for " + s;
                                        continue;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            
            StringBuilder id = new StringBuilder();
            if (namespace != null && namespace.length() > 0) {
                if (clz == null || clz.indexOf('.') == -1) {
                    id.append(namespace);
                    id.append('.');
                } // else: Don't include namespace since we're overriding it here

            }
            if (clz != null && clz.length() > 0) {
                id.append(clz);
                if (name != null) {
                    id.append('.');
                }
            }
            if (type != ElementType.CONSTRUCTOR) {
                assert name != null;
                id.append(name);
            } else {
                assert clz != null;
            }
            
            // Find a suitable place to insert it
            int idLoc = s.indexOf("* @");
            if (idLoc == -1) {
                idLoc = s.indexOf('\n');
            }
            s = s.substring(0, idLoc) + "* @id " + id.toString() + "\n" +
                    // I couldn't find a compatibility chart from YUI's main
                    // site but saw it here: http://kauriproject.org/wiki/g1/g4/g2/69-kauri.html
                    " * @compat=IE6|IE7|FF1|FF2|FF3|OPERA|SAFARI2|SAFARI3|KONQ\n" +
                    s.substring(idLoc);
            
            // Reindent
            StringBuilder reindented = new StringBuilder();
            for (String line : s.split("\n")) {
                line = line.trim();
                if (line.startsWith("*")) {
                    line = " " + line;
                }
                reindented.append(line);
                reindented.append("\n");
            }
            s = reindented.toString();
            
            result.add(s);
        }
        
        return result;
    }

    enum ElementType { FUNCTION, PROPERTY, EVENT, CONSTRUCTOR };
    
    private String fixTypos(String s) {
        // Fix typos I've found in the current YUI release (2.5.0)
        // @propery
        // @depreciated
        // @parem
        // @parm

        if (s.indexOf("@propery") != -1) {
            s = s.replaceAll("@propery", "@property");
        }
        
        if (s.indexOf("@parem") != -1) {
            s = s.replaceAll("@parem", "@param");
        }

        if (s.indexOf("@parm") != -1) {
            s = s.replaceAll("@parm", "@param");
        }

        if (s.indexOf("@depreciated") != -1) {
            s = s.replaceAll("@depreciated", "@deprecated");
        }
        
        // Normalize case:
        // @Class, @Extends, @TODO
        if (s.indexOf("@Class") != -1) {
            s = s.replaceAll("@Class", "@class");
        }
        if (s.indexOf("@Extends") != -1) {
            s = s.replaceAll("@Extends", "@extends");
        }
        if (s.indexOf("@TODO") != -1) {
            s = s.replaceAll("@TODO", "@todo");
        }
        
        return s;
    }

    private String removeLinks(String s) {
        // Remove <a ...></a> stuff
        int lt = -1;
        int lta = s.indexOf("<a");
        int ltb = s.indexOf("</a");
        if (lta == -1) {
            if (ltb == -1) {
                return s;
            } else {
                lt = ltb;
            }
        } else {
            if (ltb == -1){
                lt = lta;
            } else {
                lt = Math.min(lta, ltb);
            }
        }
        if (lt == -1) {
            return s;
        }
        
        int lte = s.indexOf('>', lt);
        if (lte == -1) {
            return s;
        }
        assert s.indexOf('<', lt+1) == -1 || s.indexOf('<', lt+1) > lte;
        
        s = s.substring(0, lt) + s.substring(lte+1);
        return removeLinks(s);
    }
    
    private String fixTypes(String s) {
        // Change type declarations from Int etc. to Number and so on
    
        
        // Some errors in the spec
        if (s.indexOf("|HTML Element}") != -1) {
            s = s.replace("|HTML Element}", "|HTMLElement}");
        }
        if (s.indexOf("{String or HTMLEle") != -1) {
            s = s.replace("{String or HTMLEle", "{String|HTMLEle");
        }
        if (s.indexOf("[int, int, int]") != -1) {
            s = s.replace("[int, int, int]", "Number[]");
        }
        if (s.indexOf("[int, int]") != -1) {
            s = s.replace("[int, int]", "Number[]");
        }
        
        int index = 0;
docParam:        
        while (index < s.length()) {
            // Find next type string
            int nextParam = s.indexOf("@param", index);
            int nextType = s.indexOf("@type", index);
            int next;
            if (nextParam == -1) {
                next = nextType;
            } else if (nextType == -1) {
                next = nextParam;
            } else {
                next = Math.min(nextParam, nextType);
            }
            if (next == -1) {
                break;
            }
            
            boolean isParam = next == nextParam;
            index = next + (isParam ? "@param".length() : "@type".length());
            
            // Iterate over all the @type params in the comma list
            while (index < s.length()) {
                
                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (c == '\n') {
                        continue docParam;
                    }
                    if (!Character.isWhitespace(c)) {
                        break;
                    }
                }

                if (isParam) {
                    // Find the opening '{'
                    if (s.charAt(index) != '{') {
                        // Nope, this @param doesn't specify the type
                        continue docParam;
                    }
                }

                int begin = index;

                // Find the end
                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (c == '}') {
                        break;
                    }
                    if (c == '\n') {
                        continue docParam;
                    }
                    if (Character.isWhitespace(c)) {
                        break;
                    }
                }
                int end = index;

                String type = s.substring(begin, end);
                String translatedType = translateType(type);
                if (translatedType != null) {
                    s = s.substring(0, begin) + translatedType + s.substring(end);
                    // Recurse
                    return fixTypes(s);
                }
                
                if (!isParam) {
                    break;
                }
                
                if (s.charAt(index) == '}') {
                    break;
                }
                
                // Skip spaces
                for (; index < s.length(); index++) {
                    char c = s.charAt(index);
                    if (!Character.isWhitespace(c)) {
                        break;
                    }
                }
                if (s.charAt(index) == '}') {
                    break;
                }
                if (s.charAt(index) == '/' || s.charAt(index) == ',' || s.charAt(index) == '|') {
                    index++;
                } else {
                    System.err.println("WARNING: Unexpected page - " +  s.charAt(index) + " in " + s.substring(index-10, index+10));
                    //assert false : s.charAt(index) + " in " + s.substring(index-10, index+10);
                    return s;
                }
                
            }
        }
        
        return s;
    }
    
    private String translateType(String s) {
        // object => Object
        // Int => Number
        // int => Number
        // string => String
        // function => Function
        // boolean => Boolean
        // float => Float

        // Not yet done:
        // Variant => ?
        // String[] => Array<String>
        
        // Not sure what to do about this:
        // @type string|{searchExp: string, replaceStr: string}
        // @type [int, int]
        // @type {"x": int, "y": int}

        if ("object".equals(s)) {
            return "Object";
        }
        if ("Int".equals(s)) {
            return "Number";
        }
        if ("int".equals(s)) {
            return "Number";
        }
        if ("string".equals(s)) {
            return "String";
        }
        if ("function".equals(s)) {
            return "Function";
        }
        if ("boolean".equals(s)) {
            return "Boolean";
        }
        if ("float".equals(s)) {
            return "Number";
        }
        
        return null;
    }

    private String getProperty(String prop, String comment) {
        int index = comment.indexOf(prop);
        if (index == -1) {
            return null;
        }

        index += prop.length(); // skip e.g. @namespace
        
        // Skip spaces
        while (index < comment.length()) {
            char c = comment.charAt(index);
            if (c == '\n') {
                return "";
            }
            if (!Character.isWhitespace(c)) {
                break;
            }
            index++;
        }
        
        StringBuilder sb = new StringBuilder();
        while (index < comment.length()) {
            char c = comment.charAt(index);
            if (Character.isWhitespace(c)) {
                break;
            }
            sb.append(c);
            index++;
        }
        
        return sb.toString();
    }
    
    enum State { INIT, SEEN_SLASH, SEEN_SLASH_STAR, IN_BLOCK_COMMENT,
        SEEN_COMMENT_STAR, IN_JSDOC_COMMENT, SEEN_JSDOC_COMMENT_STAR };
    
    private void findComments(String js, List<String> comments, Map<String,String> nextLines) {
        State state = State.INIT;
        int commentBegin = -1;
        for (int i = 0, n = js.length(); i < n; i++) {
            char c = js.charAt(i);
            
            switch (state) {
                case INIT:
                    if (c == '/') {
                        state = State.SEEN_SLASH;
                    }
                    break;
                case SEEN_SLASH:
                    if (c == '*') {
                        state = State.SEEN_SLASH_STAR;
                    } else {
                        state = State.INIT;
                    }
                    break;
                case SEEN_SLASH_STAR:
                    if (c == '*') {
                        state = State.IN_JSDOC_COMMENT;
                        commentBegin = i-2;
                    } else {
                        state = State.IN_BLOCK_COMMENT;
                    }
                    break;
                case IN_BLOCK_COMMENT:
                    if (c == '*') {
                        state = State.SEEN_COMMENT_STAR;
                    }
                    break;
                case SEEN_COMMENT_STAR:
                    if (c == '/') {
                        state = State.INIT;
                    } else if (c != '*') {
                        state = State.IN_BLOCK_COMMENT;
                    }
                    break;
                case IN_JSDOC_COMMENT:
                    if (c == '*') {
                        state = State.SEEN_JSDOC_COMMENT_STAR;
                    }
                    break;
                case SEEN_JSDOC_COMMENT_STAR:
                    if (c == '/') {
                        // TODO - pull out the comment
                        assert commentBegin != -1;
                        String comment = js.substring(commentBegin, i+1); // +1: inlude /
                        String nextLine = getNextLineStart(js, i+1);
                        if (nextLine != null) {
                            nextLines.put(comment, nextLine);
                        }
                        
                        comments.add(comment);
                        commentBegin = -1;
                        state = State.INIT;
                    } else if (c != '*') {
                        state = State.IN_JSDOC_COMMENT;
                    }
                    break;
                default:
                    assert false : state;
            }
        }
        
        //return comments;
    }
    
    private String getNextLineStart(String s, int index) {
        // skip to next line
        index = s.indexOf('\n', index)+1;
        for (int i = index; i < s.length(); i++) {
            if (!Character.isWhitespace(s.charAt(i))) {
                // Found first char
                char f = s.charAt(i);
                if (!Character.isLetter(f) && f != '_' && f != '$') {
                    return null;
                }
                for (int j = i; j < s.length(); j++) {
                    if (s.charAt(j) == '.' || s.charAt(j) == '_' || s.charAt(j) == '$') {
                        continue;
                    } else if (!Character.isLetter(s.charAt(j)) || Character.isWhitespace(s.charAt(j))) {
                        // Found end... now look for colon or equals
                        for (int k = j; k < s.length(); k++) {
                            if (!Character.isWhitespace(s.charAt(k))) {
                                if (s.charAt(k) == ':' || s.charAt(k) == '=') {
                                    String next = s.substring(i, j);
                                    if (next.startsWith("this.")) {
                                        next = next.substring("this.".length());
                                    } else if (next.startsWith("proto.")) {
                                        next = next.substring("proto.".length());
                                    }
                                    return next;
                                } else {
                                    return null;
                                }
                            }
                        }
                    }
                }
            }
        }
        
        return null;
    }
    
    private String readFile(File file) {
        StringBuilder sb = new StringBuilder();
        try {
            BufferedReader f = new BufferedReader(new FileReader(file));
            while (true) {
                String line = f.readLine();
                if (line == null) {
                    break;
                }
                sb.append(line);
                sb.append("\n");
            }
            f.close();
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(1);
        }
        
        return sb.toString();
    }
    
    
    
//    private void functionOrProperty(Document doc, Element function, Writer writer, ElementType elementType) {
//        String name = function.getAttribute("name");
//        String type = function.getAttribute("return");
//        if (type.indexOf(',') != -1) {
//            String[] types = type.split(",");
//            StringBuilder sb = new StringBuilder();
//            for (String t : types) {
//                if (sb.length() > 0) {
//                    sb.append("|");
//                }
//                sb.append(t.trim());
//            }
//            type = sb.toString();
//        }
//        assert type != null;
//        
//        String desc = null;
//        String longDesc = null;
//        String added = null;
//        String removed = null;
//        String deprecated = null;
//        List<String> paramList = new ArrayList<String>();
//        NodeList nl = function.getChildNodes();
//        String codeExample = null;
//        for (int i = 0; i < nl.getLength(); i++) {
//            Node n = nl.item(i);
//            if (n.getNodeType() != Node.ELEMENT_NODE) {
//                continue;
//            }
//            Element element = (Element)n;
//            String tag = element.getTagName();
//            if ("desc".equals(tag)) {
//                desc = getText(element);
//            } else if ("longdesc".equals(tag)) {
//                longDesc = getText(element);
//            } else if ("added".equals(tag)) {
//                added = getText(element);
//            } else if ("removed".equals(tag)) {
//                removed = getText(element);
//            } else if ("deprecated".equals(tag)) {
//                deprecated = getText(element);
//            } else if ("params".equals(tag)) {
//                String paramName = element.getAttribute("name");
//                String paramType = element.getAttribute("type");
//                String paramDesc = getText(element);
//                String param;
//                if (paramType != null) {
//                    param = "@param {" + paramType + "} " + paramName + " " + paramDesc;
//                } else {
//                    param = "@param " + paramName + " " + paramDesc;
//                }
//                paramList.add(param);
//            } else if ("example".equals(tag)) {
//                // Find the HTML element
//                NodeList enl = element.getChildNodes();
//                for (int j = 0; j < enl.getLength(); j++) {
//                    Node en = enl.item(j);
//                    if (en.getNodeType() != Node.ELEMENT_NODE) {
//                        continue;
//                    }
//                    Element html = (Element)en;
//                    //if (html.getTagName().equals("code")) {
//                    if (html.getTagName().equals("html")) {
//                        codeExample = getText(html);
//                        break;
//                    }
//                }
//            } else {
//                System.err.println("WARNING: ignoring attribute " + tag);
//            }
//        }
//
//        StringBuilder sb = new StringBuilder();
//        sb.append("/**\n");
//
//        if (desc != null) {
//            for (String s : desc.split("\n")) {
//                sb.append(" * ");
//                sb.append(s);
//                sb.append("\n");
//            }
//        }
//
//        if (longDesc != null) {
//            for (String s : longDesc.split("\n")) {
//                sb.append(" * ");
//                sb.append(s);
//                sb.append("\n");
//            }
//        }
//        
//        if (codeExample != null) {
//            sb.append(" * {@code ");
//            sb.append(codeExample);
//            sb.append("}\n");
//        }
//        
//        sb.append(" * ");
//        sb.append("@id jQuery.");
//        sb.append(name);
//        sb.append("\n");
//        
//        if (desc != null) {
//            for (String param : paramList) {
//                for (String s : param.split("\n")) {
//                    sb.append(" * ");
//                    sb.append(s);
//                    sb.append("\n");
//                }
//            }
//        }
//
//        if (type != null) {
//            sb.append(" * @type ");
//            sb.append(type);
//            sb.append("\n");
//        }
//        
//        if (elementType == ElementType.PROPERTY) {
//            sb.append(" * @property\n");
//        }
//        
//        if (added != null) {
//            sb.append(" * @since ");
//            sb.append(added);
//            sb.append("\n");
//        }
//        
//        if (removed != null) {
//            sb.append(" * @removed ");
//            sb.append(removed);
//            sb.append("\n");
//        }
//        
//        if (deprecated != null) {
//            sb.append(" * @deprecated ");
//            sb.append(deprecated);
//            sb.append("\n");
//        }
//        
//        sb.append(" */\n\n");
//        try {
//            writer.write(sb.toString());
//        } catch (IOException ex) {
//            Logger.getLogger(Main.class.getName()).log(Level.SEVERE, null, ex);
//            System.exit(0);
//        }
//    }
}

