
package org.netbeans.jsstubsgenerator.core.mdn;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import org.netbeans.jsstubsgenerator.stub.GeneratorUtils;
import org.netbeans.jsstubsgenerator.stub.Origin;
import org.netbeans.jsstubsgenerator.stub.Parameter;
import org.netbeans.jsstubsgenerator.stub.StubObject;
import org.netbeans.jsstubsgenerator.stub.Type;

/**
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class XMLHttpRequestGenerator {

    public static void generateXMLHttpRequest(Map<String, StubObject> allStubs) throws IOException {
        // general definition
        StubObject xmlHttpRequest = new StubObject(Origin.CORE, "XMLHttpRequest", StubObject.GLOBAL);
        xmlHttpRequest.setPrune(true);
        xmlHttpRequest.setType(Type.OBJECT);
        xmlHttpRequest.setSummary("XMLHttpRequest is a JavaScript object that was designed by Microsoft and adopted by Mozilla, Apple, and Google. It's now being standardized in the W3C. It provides an easy way to retrieve data at a URL. Despite its name, XMLHttpRequest can be used to retrieve any type of data, not just XML, and it supports protocols other than HTTP (including file and ftp).");
        xmlHttpRequest.setSyntax("var req = new XMLHttpRequest();");

        // methods definition
        String[][] methods = new String[][]{
            {"abort", "undefined", "Aborts the request if it has already been sent."},
            {"getAllResponseHeaders", "String", "Returns all the response headers as a string, or null if no response has been received. Note: For multipart requests, this returns the headers from the current part of the request, not from the original channel."},
            {"getResponseHeader", "String", "Returns the string containing the text of the specified header, or null if either the response has not yet been received or the header doesn't exist in the response."},
            {"open", "undefined", "Initializes a request. This method is to be used from JavaScript code; to initialize a request from native code, use openRequest()instead."},
            {"overrideMimeType", "undefined", "Overrides the MIME type returned by the server. This may be used, for example, to force a stream to be treated and parsed as text/xml, even if the server does not report it as such.This method must be called before send()."},
            {"send", "undefined", "Sends the request. If the request is asynchronous (which is the default), this method returns as soon as the request is sent. If the request is synchronous, this method doesn't return until the response has arrived."},
            {"setRequestHeader", "undefined", "Sets the value of an HTTP request header.You must call setRequestHeader() after open(), but before send()."}
        };
        List<StubObject> generatedInstanceMethods = generateInstanceMethods(methods);
        for (StubObject instanceMethod : generatedInstanceMethods) {
            xmlHttpRequest.putAttribute(instanceMethod);
        }

        // properties definition
        String[][] properties = new String[][]{
            {"onreadystatechange", "Function", "A JavaScript function object that is called whenever the readyState attribute changes. The callback is called from the user interface thread."},
            {"readyState", "Number", "The state of the request."},
            {"response", "undefined", "The response entity body according to responseType, as an ArrayBuffer, Blob, Document , JavaScript object (for \"json\"), or string. This is null if the request is not complete or was not successful."},
            {"responseText", "String", "The response to the request as text, or null if the request was unsuccessful or has not yet been sent. Read-only."},
            {"responseType", "XMLHttpRequestResponseType", "Can be set to change the response type. This tells the server what format you want the response to be in."},
            {"responseXML", "Document", "The response to the request as a DOM Document object, or null if the request was unsuccessful, has not yet been sent, or cannot be parsed as XML or HTML. The response is parsed as if it were a text/xml stream. When the responseType is set to \"document\" and the request has been made asynchronously, the response is parsed as it were a text/html stream . Read-only."},
            {"status", "Number", "The status of the response to the request. This is the HTTP result code (for example, status is 200 for a successful request). Read-only."},
            {"statusText", "String", "The response string returned by the HTTP server. Unlike status, this includes the entire text of the response message (\"200 OK\", for example). Read-only."},
            {"timeout", "Number", "The number of milliseconds a request can take before automatically being terminated. A value of 0 (which is the default) means there is no timeout."},
            {"upload", "XMLHttpRequestUpload", "The upload process can be tracked by adding an event listener to upload."},
            {"withCredentials", "Boolean", "Indicates whether or not cross-site Access-Control requests should be made using credentials such as cookies or authorization headers. The default is false."}
        };
        List<StubObject> generatedInstanceProperties = generateInstanceProperties(properties);
        for (StubObject instanceProperty : generatedInstanceProperties) {
            xmlHttpRequest.putAttribute(instanceProperty);
        }

        allStubs.put("XMLHttpRequest", xmlHttpRequest);
    }

    private static List<StubObject> generateInstanceMethods(String[][] methods) {
        List<StubObject> methodStubs = GeneratorUtils.generateMethods(methods, "XMLHttpRequest", false);
        for (StubObject method : methodStubs) {
            if ("open".equals(method.getName())) {
                method.setSyntax("xmlHttpRequest.open(method,url,async,uname,pswd)");
                method.addParameter(new Parameter("method", "String", "The HTTP method to use, such as \"GET\", \"POST\", \"PUT\", \"DELETE\", etc. Ignored for non-HTTP(S) URLs."));
                method.addParameter(new Parameter("url", "String", "The URL to which to send the request."));
                method.addParameter(new Parameter("async", "Boolean", "An optional boolean parameter, defaulting to true, indicating whether or not to perform the operation asynchronously. If this value is false, the send()method does not return until the response is received. If true, notification of a completed transaction is provided using event listeners. This must be true if the multipart attribute is true, or an exception will be thrown.", true));
                method.addParameter(new Parameter("user", "String", "The optional user name to use for authentication purposes; by default, this is an empty string.", true));
                method.addParameter(new Parameter("password", "String", "The optional password to use for authentication purposes; by default, this is an empty string.", true));
            } else if ("setRequestHeader".equals(method.getName())) {
                method.setSyntax("xmlHttpRequest.setRequestHeader()");
                method.addParameter(new Parameter("header", "String", "The name of the header whose value is to be set."));
                method.addParameter(new Parameter("value", "String", "The value to set as the body of the header."));
            } else if ("getResponseHeader".equals(method.getName())) {
                method.setSyntax("xmlHttpRequest.getResponseHeader()");
                method.addParameter(new Parameter("header", "String", ""));
            } else if ("overrideMimeType".equals(method.getName())) {
                method.setSyntax("xmlHttpRequest.overrideMimeType()");
                method.addParameter(new Parameter("mimetype", "String", ""));
            } else if ("send".equals(method.getName())) {
                method.setSyntax("xmlHttpRequest.send(string)");
                method.addParameter(new Parameter("string", "String", "Sends the request off to the server."));
            } else {
                method.setSyntax("xmlHttpRequest." + method.getName() + "()");
            }
        }
        return methodStubs;
    }

    private static List<StubObject> generateInstanceProperties(String[][] properties) {
        List<StubObject> propertyStubs = GeneratorUtils.generateProperties(Origin.CORE, properties, "XMLHttpRequest", false);
        return propertyStubs;
    }
}
