package org.netbeans.jsstubsgenerator.dom;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.netbeans.jsstubsgenerator.stub.Origin;
import org.netbeans.jsstubsgenerator.stub.Parameter;
import org.netbeans.jsstubsgenerator.stub.Return;
import org.netbeans.jsstubsgenerator.stub.StubObject;
import org.netbeans.jsstubsgenerator.stub.Type;

/**
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class DomGenerator {

    private static final Logger LOGGER = Logger.getLogger(DomGenerator.class.getName());

    public static void parseDomDocs(Map<String, StubObject> stubs, File dirWithDomStubs) throws IOException {
        if (!dirWithDomStubs.exists()) {
            LOGGER.log(Level.WARNING, "Dir with dom docs {0} doesn''t exist!", dirWithDomStubs.getPath());
        }

        // load and parse stubs from W3C files
        loadStubs(dirWithDomStubs, stubs);

        // updates stubs where it was not possible to parse it automatically
        updateStubsManually(stubs);

        // update stubs according implement clause of stub objects
        updateAccordingImplementClause(stubs);
    }

    private static void updateStubsManually(Map<String, StubObject> allStubs) {
        CustomUpdater.updateHTMLElementObject(allStubs.get("HTMLElement"));
        CustomUpdater.updateElementObject(allStubs.get("Element"));
        CustomUpdater.updateDocumentObject(allStubs.get("Document"));
    }

    private static void loadStubs(File dirWithDomStubs, Map<String, StubObject> allStubs) throws IOException {
        // TODO - do we want to parse DOM1 stubs?

        // load dom2 stubs
        processDomDocs(dirWithDomStubs.listFiles(new DomJsFileFilters.Dom2()), allStubs);

        // load dom3 stubs
        processDomDocs(dirWithDomStubs.listFiles(new DomJsFileFilters.Dom3()), allStubs);

        // enhance for inheritance
        Inheritance.enhanceForInheritance(allStubs);
    }

    private static void processDomDocs(File[] docFiles, Map<String, StubObject> allStubs) throws IOException {
        for (File docFile : docFiles) {
            Document doc = Jsoup.parse(docFile, "UTF-8");

            // find all elements definitions
            Elements allDls = doc.select("dl");
//            System.out.println("Found DOM elements, in file = " + docFile.getName() + " =====================");
            for (Element anyDl : allDls) {
                Elements idls = anyDl.select("dd > dl:contains(IDL Definition)");
                for (Element idl : idls) {
                    Element elDefinition = idl.parent().previousElementSibling();
                    String attributeName = getCleanedName(elDefinition.text());
                    StubObject stubObject = allStubs.get(attributeName);
                    if (stubObject == null) {
                        stubObject = new StubObject(Origin.DOM, attributeName, getParent(attributeName));
                    } else {
                        stubObject.setParent(getParent(attributeName));
                    }
                    stubObject.setType(Type.OBJECT);
                    stubObject.setSummary(getSummaryInfo(elDefinition));
                    for (Map.Entry<String, StubObject> entry : getDomAttributes(attributeName, elDefinition.nextElementSibling()).entrySet()) {
                        stubObject.putAttribute(entry.getValue());
                    }
                    allStubs.put(attributeName, stubObject);
                    //                    System.out.println(getCleanedName(element) + " for " + element);
                }
            }
        }
    }

    private static String getCleanedName(String nameHeader) {
        String withoutAvailabilityNote = nameHeader.indexOf("(") == -1 ? nameHeader.trim() : nameHeader.substring(0, nameHeader.indexOf("(")).trim();
        // there should be no "introduced in ..." string
        int indexOfIntroduced = withoutAvailabilityNote.indexOf("introduced in");
        if (indexOfIntroduced != -1) {
            withoutAvailabilityNote = withoutAvailabilityNote.substring(0, indexOfIntroduced).trim();
        }
        String[] splitedString = withoutAvailabilityNote.split(" ");
        return splitedString[splitedString.length - 1];
    }

    private static String getSummaryInfo(Element elDefinition) {
        StringBuilder summarySB = new StringBuilder();
        Elements descriptions = elDefinition.nextElementSibling().select("p");
        for (Element description : descriptions) {
            if (description.parent() == elDefinition.nextElementSibling()) {
                summarySB.append(description.text());
            }
        }
//        System.err.println("=====================================");
//        System.err.println(getCleanedName(elDefinition.text()));
//        System.err.println(summarySB.toString());
        return summarySB.toString();
    }

    private static Map<String, StubObject> getDomAttributes(String parent, Element baseElement) {
        Map<String, StubObject> attributes = new HashMap<String, StubObject>();
//        System.err.println("==============");
//        System.err.println(parent);
        Elements subsections = baseElement.select("dt");
        for (Element subsection : subsections) {
            String subsectionText = subsection.text().trim();
            if ("Defined Constants".equals(subsectionText)) {
                Element constWrapper = subsection.nextElementSibling();
                for (Element constNameElement : constWrapper.select("dt")) {
                    Element constDescElement = constNameElement.nextElementSibling();
                    String constName = constNameElement.text().split(" ")[0];
                    if (constName.endsWith(",")) {
                        constName = constName.substring(0, constName.indexOf(","));
                    }
                    String constDesc = constDescElement.text().trim();
                    StubObject constant = new StubObject(Origin.DOM, constName, parent);
                    constant.setType(Type.PROPERTY);
                    constant.setSummary(constDesc);
                    // TODO - returns could be get from IDL definition - all DOM consts look to be of type Number
                    constant.setReturns(new Return("Number", null));
                    constant.setStatic(true);
                    constant.setSyntax(parent + "." + constName);
                    attributes.put(constName, constant);
                    if (!constName.toUpperCase().equals(constName)) {
                        LOGGER.log(Level.WARNING, "Lowercase constant={0}, parent={1}", new Object[]{constName, parent});
                    }
//                    System.err.println(constName + " - " + constDesc);
                }
            } else if ("Methods".equals(subsectionText)) {
                Element methodsWrapper = subsection.nextElementSibling();
                for (Element methodNameElement : methodsWrapper.select("dt > code[class=method-name]")) {
                    Element methodDescElement = methodNameElement.parent().nextElementSibling();
                    String methodHeader = methodNameElement.parent().text().trim();
                    String methodName = methodHeader.split(" ")[0];
                    String methodDesc = getMethodDesc(methodDescElement);
                    StubObject method = new StubObject(Origin.DOM, methodName, parent);
                    method.setType(Type.METHOD);
                    method.setSummary(methodDesc);
                    for (Parameter parameter : getMethodParameters(methodDescElement)) {
                        method.addParameter(parameter);
//                        System.err.println(parameter.getName()+"-"+parameter.getType()+":"+parameter.getDesc());
                    }
                    method.setReturns(getMethodReturns(methodDescElement));
                    StringBuilder syntaxBuilder = new StringBuilder(getStringWithFirstLowercased(parent));
                    syntaxBuilder.append(".").append(methodName).append("(");
                    String delim = "";
                    for (Parameter param : method.getParameters()) {
                        syntaxBuilder.append(delim).append(param.getName());
                        delim = ", ";
                    }
                    syntaxBuilder.append(")");
                    method.setSyntax(syntaxBuilder.toString());
                    attributes.put(methodName, method);
                    if (methodName.indexOf(" ") != -1) {
                        LOGGER.log(Level.WARNING, "Method name contain space: method={0}, parent={1}", new Object[]{methodName, parent});
                    }
//                    System.err.println(methodName + "{" + getMethodReturns(methodDescElement).getType() + "} - " + methodDesc);
                }
            } else if ("Attributes".equals(subsectionText)) {
                Element attrsWrapper = subsection.nextElementSibling();
                for (Element attrNameElement : attrsWrapper.select("dt")) {
                    Element attrDescElement = attrNameElement.nextElementSibling();
                    String attrHeader = attrNameElement.text().trim();
                    String attrName = attrHeader.split(" ")[0];
                    String attrDesc = attrDescElement.text().trim();
                    String attrReturns = getReturns(attrHeader);
                    StubObject attribute = new StubObject(Origin.DOM, attrName, parent);
                    attribute.setType(Type.PROPERTY);
                    attribute.setSummary(attrDesc);
                    attribute.setReturns(new Return(attrReturns, null));
                    attribute.setSyntax(getStringWithFirstLowercased(parent) + "." + attrName);
                    attributes.put(attrName, attribute);
//                    System.err.println(attrName + "{" + attrReturns + "} - " + attrDesc);
                }
            }
//            System.err.println(subsectionText);
        }

        return attributes;
    }

    private static String getStringWithFirstLowercased(String string) {
        return string.substring(0, 1).toLowerCase() + string.substring(1);
    }

    private static String getReturns(String attrHeader) {
        int indexOf = attrHeader.indexOf("of type ");
        if (indexOf != -1) {
            String tmp = attrHeader.substring(indexOf + 8);
            String returnText = (tmp.indexOf(",") != -1) ? tmp.substring(0, tmp.indexOf(",")) : tmp;
            return guessReturns(returnText);
        }
        return null;
    }

    private static String guessReturns(String returnText) {
        if ("DOMString".equals(returnText)) {
            return "String";
        } else if (returnText.contains("unsigned") || "long".equals(returnText) || "short".equals(returnText)) {
            return "Number";
        } else if ("boolean".equals(returnText)) {
            return "Boolean";
        } else {
            return returnText;
        }
    }

    private static String getMethodDesc(Element htmlEl) {
        if (htmlEl.html().indexOf("<div", 1) == -1) {
            return htmlEl.html();
        }
        String htmlSubstring = htmlEl.html().substring(0, htmlEl.html().indexOf("<div", 1));
        String endingSeq = htmlSubstring.substring(htmlSubstring.length() - 15, htmlSubstring.length() - 5);
        if (endingSeq.contains(">")) {
            // probably contains HTML, return the original substring
            return cleanHtmlFromText(htmlSubstring);
        } else {
            // otherwise clean the html as much as possible
            int index = htmlEl.text().indexOf(endingSeq);
            if (index == -1) {
                return cleanHtmlFromText(htmlSubstring);
            } else {
                return cleanHtmlFromText(htmlEl.text().substring(0, index + 12));
            }
        }
    }

    private static String cleanHtmlFromText(String htmlText) {
        return Jsoup.parse(htmlText).text().trim();
    }

    private static List<Parameter> getMethodParameters(Element methodDescElement) {
        List<Parameter> parameters = new LinkedList<Parameter>();
        Elements parameterDivs = methodDescElement.select("div[class=parameters]");
        if (!parameterDivs.isEmpty()) {
            Element parameterDiv = parameterDivs.get(0);
            for (Element paramNameElement : parameterDiv.select("dt")) {
                String paramNameHeader = paramNameElement.text();
                String paramName = paramNameHeader.split(" ")[0];
                String paramReturns = getReturns(paramNameHeader);
                String paramDesc = paramNameElement.nextElementSibling().text();
                parameters.add(new Parameter(paramName, paramReturns, paramDesc));
            }
        }
        return parameters;
    }

    private static Return getMethodReturns(Element methodDescElement) {
        Elements returnDivs = methodDescElement.select("div[class=return]");
        if (!returnDivs.isEmpty()) {
            Element returnDiv = returnDivs.get(0);
            Elements tds = returnDiv.select("tr > td");
            if (!tds.isEmpty()) {
                Element returnTd = tds.get(0);
                String returnName = returnTd.text();
                String returnDesc = returnTd.nextElementSibling().text();
                return new Return(guessReturns(returnName), returnDesc);
            }
        }
        return new Return("undefined", null);
    }

    private static String getParent(String object) {
        if ("HTMLDocument".equals(object)) {
            return "Document";
        } else if ("Document".equals(object)) {
            return "Node";
        } else if ("HTMLElement".equals(object)) {
            return "Element";
        } else if ("Element".equals(object)) {
            return "Node";
        }
        return StubObject.GLOBAL;
    }

    private static void updateAccordingImplementClause(Map<String, StubObject> stubs) {
        for (Map.Entry<String, StubObject> stubsEntry:  stubs.entrySet()) {
            StubObject so = stubsEntry.getValue();
            for (String implement : so.getImplements()) {
                StubObject implementedClass = stubs.get(implement);
                for (Map.Entry<String, StubObject> attrEntry : implementedClass.getAttributes().entrySet()) {
                    try {
                        // change parent of the object
                        StubObject newAttribute = attrEntry.getValue().clone();
                        newAttribute.setParent(so.getName());
                        so.putAttribute(newAttribute);
                    } catch (CloneNotSupportedException ex) {
                        Logger.getLogger(DomGenerator.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            }
        }
    }
}
