/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * Oracle and Java are registered trademarks of Oracle and/or its affiliates.
 * Other names may be trademarks of their respective owners.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 *
 * Portions Copyrighted 2015 Denis Anisimov
 */
package org.netbeans.server.uihandler.statistics;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.LogRecord;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import javax.servlet.jsp.PageContext;
import org.netbeans.server.uihandler.Statistics;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author denis
 */
@ServiceProvider(service = Statistics.class)
public class Vaadin extends Statistics<Vaadin.VaadinData>{
    
    public static final String PROJECT_WIZARDS = "vaadinProjectWizards";
    public static final String COMPONENT_WIZARDS = "vaadinComponentWizards";
    public static final String ACTIONS = "vaadinProjectActions";
    public static final String ADDONS ="vaadinAddons";
    public static final String FIXES ="vaadinFixes";
    
    public static final String USAGE = "usage";
    public static final String CATEGORY = "category";
    
    private static final String UI_FIX_PACKAGE_NAME = "org.vaadin.netbeans.editor.analyzer"; // NOI18N
    
    private static final String UI_ACTION_PACKAGE_NAME = "org.vaadin.netbeans.maven.project"; // NOI18N
    
    private static final String UI_PROJECT_WIZARD_PACKAGE_NAME = "org.vaadin.netbeans.maven.ui.wizard"; // NOI18N
    
    private static final String UI_COMPONENT_WIZARD_PACKAGE_NAME = "org.vaadin.netbeans.ui.wizard"; // NOI18N
    
    private static final String UI_POM_UTILS_PACKAGE_NAME = "org.vaadin.netbeans.utils"; // NOI18N
    
    
    private static final String IDS_PREFS = "IDS";// NOI18N
    private static final String ADD_ONS_PREFS = "ADD_ONS";// NOI18N
    
    private static final VaadinData EMPTY = new VaadinData();
   
    public enum Category {
        ALL, ADD_ONS, PROJECTS, WIDGETS, ACTIONS, HINTS;
    }
    
    public enum UsageFilter {
        ALL, UNIQUE, OVERALL;
    }
    
    enum DataType {
        FIX, ACTION, PROJECT, COMPONENT;
    }
   
    @Override
    protected VaadinData newData() {
        return EMPTY;
    }

    @Override
    protected VaadinData process(LogRecord rec) { 
        String recordBundleName = rec.getResourceBundleName();
        // rec (received after upload gestures) doesn't contain logger name (WTH!!!)
        // so we have to rely on bundle name which is package of the class whose 
        // logger is used to report the record
        if (UI_FIX_PACKAGE_NAME.equals(recordBundleName)){
            return new VaadinData(DataType.FIX, new StatData(rec.getMessage()));
        }
        else if (UI_ACTION_PACKAGE_NAME.equals(recordBundleName)){
            return new VaadinData(DataType.ACTION, new StatData(rec.getMessage()));
        }
        else if (UI_PROJECT_WIZARD_PACKAGE_NAME.equals(recordBundleName)){
            return new VaadinData(DataType.PROJECT, new StatData(rec.getMessage()));
        }
        else if (UI_COMPONENT_WIZARD_PACKAGE_NAME.equals(recordBundleName)){
            return new VaadinData(DataType.COMPONENT, new StatData(rec.getMessage()));
        }
        else if (UI_POM_UTILS_PACKAGE_NAME.equals(recordBundleName)){
            /*
             * Add-ons are reported either via code completion or via 
             * Add-on browser. Both actions are handled by POMUtils class in 
             * org.vaadin.netbeans.utils package.
             *
             * TODO: artifactId and version are not sent with statistics at the 
             * moment, so only overall number of addons is collected here. But 
             * more sophisticated logic (and data type) can be used here once 
             * plugin starts to send mentioned data.
             */
            return new VaadinData(1);
        }
        return EMPTY;
    }

    @Override
    protected VaadinData finishSessionUpload(String userId, int sessionNumber, boolean initialParse, VaadinData data) {
        if ( data == null ){
            return null;
        }
        if ( data == EMPTY){
            return EMPTY;
        }
        assert data.getIds().isEmpty() || (data.getIds().size()==1 && data.getIds().iterator().next().equals(userId));
        int addons = 0;
        if ( !data.getAddons().isEmpty()){
            addons = data.getAddons().values().iterator().next();
        }
        return new VaadinData(userId, makeDataMap(data), addons);
    }

    @Override
    protected VaadinData join(VaadinData one, VaadinData two) {
        if ( one == EMPTY ){
            return two;
        }
        else if (two == EMPTY ){
            return one;
        }
        Set<String> oneIds = one.getIds();
        Set<String> twoIds = two.getIds();
        
        Set<String> ids = new HashSet<String>(oneIds);
        ids.addAll(twoIds);
        
        Map<String, StatData> actions = mergeStatData(one.getActions(), two.getActions());
        Map<String, StatData> components= mergeStatData(one.getComponents(), two.getComponents());
        Map<String, StatData> fixes = mergeStatData(one.getFixes(), two.getFixes());
        Map<String, StatData> projects= mergeStatData(one.getProjects(), two.getProjects());
        
        Map<DataType,Map<String, StatData>> map = new EnumMap<DataType, Map<String, StatData>>(DataType.class);
        map.put(DataType.FIX, fixes);
        map.put(DataType.ACTION, actions);
        map.put(DataType.COMPONENT, components);
        map.put(DataType.PROJECT, projects);
        
        return new VaadinData(ids,map,  mergeAmounts(one.getAddons(), two.getAddons()));
    }

    @Override
    protected void write(Preferences pref, VaadinData data) throws BackingStoreException {
        if ( data==null){
            return;
        }
        List<String> ids = new ArrayList<String>(data.getIds());
        storeIds(pref.node(IDS_PREFS), ids);
        Map<String, Integer> backMap = new HashMap<String, Integer>();
        Map<String, Integer> addons = new  HashMap<String, Integer>();
        for(int i=0; i<ids.size(); i++){
            backMap.put(ids.get(i), i);
            addons.put(Integer.toString(i), data.getAddons().get(ids.get(i)));
        }
        writeStatData(DataType.ACTION, data.getActions(), pref, backMap);
        writeStatData(DataType.COMPONENT, data.getComponents(), pref, backMap);
        writeStatData(DataType.FIX, data.getFixes(), pref, backMap);
        writeStatData(DataType.PROJECT, data.getProjects(), pref, backMap);
        storeMap(pref.node(ADD_ONS_PREFS), addons);
    }

    @Override
    protected VaadinData read(Preferences pref) throws BackingStoreException {
        Map<String, String> ids = readIds(pref.node(IDS_PREFS));
        
        Map<DataType,Map<String, StatData>> map = new EnumMap<DataType, Map<String, StatData>>(DataType.class);
        for(DataType type: DataType.values()){
            map.put(type, getData(type, pref, ids));
        }
        
        return new VaadinData(new HashSet<String>(ids.values()), map, readAddons(pref, ids));
    }

    @Override
    protected void registerPageContext(PageContext page, String name, VaadinData data) {
        super.registerPageContext(page, name, data); 
        setProjectsAttribute(page, data);
        setComponentsAttribute(page, data);
        setActionsAttribute(page, data);
        setAddonsAttribute(page, data);
        setFixesAttribute(page, data);
        
        Object cat = page.getAttribute(CATEGORY, PageContext.REQUEST_SCOPE); // NOI18N
        if ( cat == null ){
            page.setAttribute(CATEGORY, Category.ALL.name());
        }
        else {
            page.setAttribute(CATEGORY, cat);
        }
        page.setAttribute("categories", Category.values());
        
        Object usage = page.getAttribute(USAGE, PageContext.REQUEST_SCOPE); // NOI18N
        if ( cat == null ){
            page.setAttribute(USAGE, UsageFilter.ALL.name());
        }
        else {
            page.setAttribute(USAGE, usage);
        }
        page.setAttribute("usages", UsageFilter.values());
    }
    
    private void setAddonsAttribute(PageContext page, VaadinData data){
        Map<String, Integer> addons = data.getAddons();
        long overall=0;
        int count =0;
        for(Integer amount: addons.values()){
            overall+=amount;
            if ( amount >0 ){
                count++;
            }
        }
        page.setAttribute(ADDONS, new AddonsStat(count, overall));
    }
    
    private void setProjectsAttribute(PageContext page, VaadinData data){
        Map<String,Integer> uniqueUsers = new HashMap<String, Integer>();
        Map<String,Long> overall = new HashMap<String, Long>();
        collectUsers(data.getProjects(), overall, uniqueUsers);
        page.setAttribute(PROJECT_WIZARDS, new ProjectsStat(uniqueUsers, overall));
    }
    
    private void setFixesAttribute(PageContext page, VaadinData data){
        Map<String,Integer> uniqueUsers = new HashMap<String, Integer>();
        Map<String,Long> overall = new HashMap<String, Long>();
        collectUsers(data.getFixes(), overall, uniqueUsers);
        page.setAttribute(FIXES, new FixStat(uniqueUsers, overall));
    }
    
    private void setComponentsAttribute(PageContext page, VaadinData data){
        Map<String,Integer> uniqueUsers = new HashMap<String, Integer>();
        Map<String,Long> overall = new HashMap<String, Long>();
        collectUsers(data.getComponents(), overall, uniqueUsers);
        page.setAttribute(COMPONENT_WIZARDS, new ComponentWizardStat(uniqueUsers, overall));
    }

    private void setActionsAttribute(PageContext page, VaadinData data){
        Map<String,Integer> uniqueUsers = new HashMap<String, Integer>();
        Map<String,Long> overall = new HashMap<String, Long>();
        collectUsers(data.getActions(), overall, uniqueUsers);
        page.setAttribute(ACTIONS, new ActionStat(uniqueUsers, overall));
    }
    
    private void collectUsers(Map<String, StatData> data, Map<String, Long> overall, Map<String, Integer> uniqueUsers) {
        for (StatData stat : data.values()) {
            for (Map.Entry<String, Integer> entry : stat.getData().entrySet()) {
                Long amount = overall.get(entry.getKey());
                if ( amount ==null ){
                    amount =0L;
                }
                overall.put(entry.getKey(), entry.getValue()+amount);
                Integer count = uniqueUsers.get(entry.getKey());
                if ( count ==null ){
                    count =0;
                }
                if ( entry.getValue() >0 ){
                    count++;
                }
                uniqueUsers.put(entry.getKey(), count);
            }
        }
    }
      
    private Map<String, StatData> getData(DataType type, Preferences pref, Map<String, String> ids) throws BackingStoreException{
        Preferences typePrefs = pref.node(type.name());
        String[] indeces = typePrefs.childrenNames();
        Map<String, StatData> map = new HashMap<String, StatData>();
        for (String id : indeces) {
            map.put(ids.get(id),new StatData(readMap(typePrefs.node(id))));
        }
        return map;
    }
    
    private void writeStatData(DataType type, Map<String, StatData> map, 
            Preferences pref, Map<String,Integer> ids) throws BackingStoreException{
        pref.node(type.name()).removeNode();
        Preferences node = pref.node(type.name());
        for(Map.Entry<String,StatData> entry: map.entrySet()){
            Integer id = ids.get(entry.getKey());
            storeMap(node.node(Integer.toString(id)), entry.getValue().getData());
        } 
    }
    
    private Map<String, Integer> readMap(Preferences prefs) throws BackingStoreException{
        String[] keys = prefs.keys();
        Map<String, Integer> map = new HashMap<String, Integer>();
        for (String key : keys) {
            map.put(key, prefs.getInt(key, 0));
        }
        return map;
    }
    
    private Map<String, Integer> readAddons(Preferences prefs, Map<String, String> ids) throws BackingStoreException{
        String[] keys = prefs.node(ADD_ONS_PREFS).keys();
        Map<String, Integer> map = new HashMap<String, Integer>();
        for (String key : keys) {
            map.put(ids.get(key), prefs.getInt(key, 0));
        }
        return map;
    }
    
    private Map<String, String> readIds(Preferences prefs) throws BackingStoreException{
        String[] keys = prefs.keys();
        Map<String, String> map = new HashMap<String, String>();
        for (String key : keys) {
            map.put(key, prefs.get(key,""));
        }
        return map;
    }
    
    private void storeMap(Preferences node, Map<String, Integer> map) throws BackingStoreException {
        if (map == null){
            return;
        }
        node.clear();
        for (Map.Entry<String, Integer> entry : map.entrySet()) {
            Integer value = entry.getValue();
            if ( value != null){
                node.putInt(entry.getKey(), entry.getValue());
            }
        }
    }
    
    private void storeIds(Preferences node, List<String> list) throws BackingStoreException {
        if (list == null){
            return;
        }
        node.clear();
        int i=0;
        for (String id: list) {
            node.put(Integer.toString(i), id);
            i++;
        }
    }
    
    private Map<String, StatData> mergeStatData(Map<String, StatData> data1, Map<String, StatData> data2){
        Map<String, StatData> result = new HashMap<String, StatData>(data2);
        for(Map.Entry<String,StatData> entry: data1.entrySet()){
            result.put(entry.getKey(), 
                entry.getValue().merge(result.get(entry.getKey())));   
        }
        return result;
    }
    
    private Map<DataType,StatData> makeDataMap(VaadinData data){
        Map<DataType,StatData> map = new EnumMap<DataType, StatData>(DataType.class);
        if ( !data.getActions().isEmpty()){
            map.put(DataType.ACTION, data.getActions().values().iterator().next());
        }
        if (!data.getComponents().isEmpty()){
            map.put(DataType.COMPONENT, data.getComponents().values().iterator().next());
        }
        if (!data.getFixes().isEmpty()){
            map.put(DataType.FIX, data.getFixes().values().iterator().next());
        }
        if (!data.getProjects().isEmpty()){
            map.put(DataType.PROJECT ,data.getProjects().values().iterator().next());
        }
        return map;
    }
    
    private static Map<String,Integer> mergeAmounts(Map<String,Integer> one, Map<String,Integer> two){
        Map<String, Integer> result = new HashMap<String, Integer>(two);
        for(Map.Entry<String,Integer> entry: one.entrySet()){
            Integer secondAmount = result.get(entry.getKey());
            if ( secondAmount == null){
                secondAmount = 0;
            }
            result.put(entry.getKey(), 
                entry.getValue() +secondAmount);   
        }
        return result;
    }
    
    public class AddonsStat {
        private AddonsStat(int uniqueUsers, long overall){
            myUniqueUsers = uniqueUsers;
            myOverall = overall;
        }
        
        public int getUniqueUsers(){
            return myUniqueUsers;
        }
        
        public long getOverallUsage(){
            return myOverall;
        }
        
        private final int myUniqueUsers;
        private final long myOverall;
    }
    
    public static class FixStat extends WebStat {
        private FixStat(Map<String,Integer> uniqueUsers, Map<String,Long> overall){
            super(uniqueUsers, overall);
        }
        
        @Override
        public String getDisplayName(String key) {
            if ( "UI_LogAddSourcePath".equals(key)){
                return "Add source path element fix";
            }else if ("UI_LogCreateGwtModule".equals(key)){
                return "Create GWT module fix";
            }else if ("UI_LogMoveClientClass".equals(key)){
                return "Move class to client pkg fix";
            } else if ("UI_LogRenamePackage".equals(key)){
                return "Rename pkg fix";
            }else if ("UI_LogGenClientRpcProxyInterface".equals(key)){
                return "Generate Client Rpc Proxy fix";
            }else if ("UI_LogCreateClientRpc".equals(key)){
                return "Generate Client Rpc Interface fix";
            }else if ("UI_LogCreateServerRpc".equals(key)){
                return "Generate Server Rpc Interface fix";
            }else if ("UI_LogGenServerRpcProxyInterface".equals(key)){
                return "Generate Server Rpc Proxy fix";
            }else if ("UI_LogClientAcceptCriterion".equals(key)){
                return "Generate Client Accept Criterion fix";
            }else if ("UI_LogGenerateConnector".equals(key)){
                return "Generate Connector fix";
            }else if ("UI_LogGenStateAccessor".equals(key)){
                return "Generate State accessors fix";
            }else if ("UI_LogSetWidgetsetWebServlet".equals(key)){
                return "Set widgetset in WebServlet annotation fix";
            }else if ("UI_LogSetUIWebServlet".equals(key)){
                return "Set UI in Servlet annotation fix";
            }else if ("UI_LogConnectToComponent".equals(key)){
                return "Associate Connector with Component fix";
            }else if ("UI_LogDeriveServletClass".equals(key)){
                return "Extends Vaadin servlet fix";
            }else if ("UI_LogSetVoidReturn".equals(key)){
                return "Set Void return type fix";
            }else if ("UI_LogSetWidgetsetForVaadinServlet".equals(key)){
                return "Set widgetset in VaadinServletConfiguration fix";
            }else if ("UI_LogUseTheme".equals(key)){
                return "Use Theme fix";
            }
            return key;
        }
    }
    
    public static class ProjectsStat extends WebStat{
        private ProjectsStat(Map<String,Integer> uniqueUsers, Map<String,Long> overall){
            super(uniqueUsers, overall);
        }

        @Override
        public String getDisplayName(String key) {
            if ( "UI_LogCreateTouchkitProject".equals(key)){
                return "Touchkit project Wizard";
            }else if ("UI_LogCreateAddOnProject".equals(key)){
                return "Add On project Wizard";
            }else if ("UI_LogCreateExampleProject".equals(key)){
                return "Example project Wizard";
            } else if ("UI_LogCreateWebProject".equals(key)){
                return "Simple Web project Wizard";
            }else if ("UI_LogCreateMultiProject".equals(key)){
                return "Multi module project Wizard";
            }else if ("UI_LogCreateSampleProject".equals(key)){
                return "Samle project Wizard (Samples category)";
            }
            return key;
        }
    }
    
    public static class ComponentWizardStat extends WebStat{
        private ComponentWizardStat(Map<String,Integer> uniqueUsers, Map<String,Long> overall){
            super(uniqueUsers, overall);
        }

        @Override
        public String getDisplayName(String key) {
            if ( "UI_LogConnectorOnly".equals(key)){
                return "Vaadin Connector Wizard";
            }else if ("UI_LogExtension".equals(key)){
                return "Vaadin Extension Wizard";
            }else if ("UI_LogFullFledged".equals(key)){
                return "Vaadin Full Fledged Widget Wizard";
            } 
            return key;
        }
    }
    
    public static class ActionStat extends WebStat{
        private ActionStat(Map<String,Integer> uniqueUsers, Map<String,Long> overall){
            super(uniqueUsers, overall);
        }

        @Override
        public String getDisplayName(String key) {
            if ( "UI_LogDebugJettyAction".equals(key)){
                return "Debug Jetty Action";
            }else if ("UI_LogRunJettyDevModeAction".equals(key)){
                return "Run Jetty with Dev Mode Action";
            }else if ("UI_LogRunJettyAction".equals(key)){
                return "Run Jetty Action";
            } else if ("UI_LogAddonBrowserAction".equals(key)){
                return "Add-On Browser Action";
            } else if ("UI_LogCompileWidgetsetAction".equals(key)){
                return "Compile Widgetset Action";
            } else if ("UI_LogCompileThemeAction".equals(key)){
                return "Compile Theme Action";
            } else if ("UI_LogDebugDevModeAction".equals(key)){
                return "Debug Dev Mode Action";
            } else if ("UI_LogRunDevModeAction".equals(key)){
                return "Run Dev Mode Action";
            } else if ("UI_LogSuperDevModeAction".equals(key)){
                return "Run SuperDev Mode Action";
            } 
            return key;
        }
    }
    
    public abstract static class WebStat {
        private WebStat(Map<String,Integer> uniqueUsers, Map<String,Long> overall){
            myUnique = Collections.unmodifiableMap(uniqueUsers);
            myOverall = Collections.unmodifiableMap(overall);
        }
        
        public Map<String,Integer> getUniqueUsers(){
            return myUnique;
        }
        
        public Map<String,Long> getOverallUsage(){
            return myOverall;
        }
        
        public abstract String getDisplayName(String key);
        
        public Map<String,Number> getDisplayPresentation(Map<String,? extends Number> map){
            Map<String,Number> result = new HashMap<String, Number>();
            for(Map.Entry<String,? extends Number> entry : map.entrySet()){
                result.put(getDisplayName(entry.getKey()), entry.getValue());
            }
            return result;
        }
        
        private final Map<String,Integer> myUnique;
        private final Map<String,Long> myOverall;
    }
    
    public static class VaadinData {
        
        VaadinData(Set<String> ids, Map<DataType,Map<String, StatData>> statData,
                Map<String,Integer> addons){            
            myIds = Collections.unmodifiableSet(ids);
            myStatData = Collections.unmodifiableMap(statData);
            myAddons = Collections.unmodifiableMap(addons);
        }
               
        private VaadinData(){            
            myIds = Collections.emptySet();
            myStatData = Collections.emptyMap();
            myAddons = Collections.emptyMap();
        }
        
        private VaadinData(DataType type , StatData data){            
            myIds = Collections.emptySet();
            myStatData = Collections.singletonMap(type, Collections.singletonMap("", data));
            myAddons = Collections.emptyMap();
        }
        
        private VaadinData(Integer addonsCount){            
            myIds = Collections.emptySet();
            myStatData = Collections.emptyMap();
            myAddons = Collections.singletonMap("", addonsCount);
        }
        
        private VaadinData(String id, Map<DataType,StatData> statData,
                Integer addons){            
            myIds = Collections.singleton(id);
            myAddons = Collections.singletonMap(id, addons);
            Map<DataType,Map<String, StatData>> data = new EnumMap<DataType, Map<String, StatData>>(DataType.class);
            for(Map.Entry<DataType,StatData> entry: statData.entrySet()){
                data.put(entry.getKey(), Collections.singletonMap(id, entry.getValue()));
            }
            myStatData = Collections.unmodifiableMap(data);
        }
        
        public Set<String> getIds(){
            return myIds;
        }
        
        public Map<String,StatData> getFixes(){
            return getData(DataType.FIX);
        }
        
        public Map<String,StatData> getActions(){
            return getData(DataType.ACTION);
        }
        
        public Map<String,StatData> getProjects(){
            return getData(DataType.PROJECT);
        }
        
        public Map<String,StatData> getComponents(){
            return getData(DataType.COMPONENT);
        }
        
        public Map<String, Integer> getAddons(){
            return myAddons;
        } 
        
        private Map<String, StatData> getData(DataType type){
            Map<String, StatData> data = myStatData.get(type);
            if ( data == null ){
                return Collections.emptyMap();
            }
            return data;
        }
        
        private final Set<String> myIds;
        
        private final Map<DataType,Map<String, StatData>> myStatData;
        
        private final Map<String,Integer> myAddons;
    }
    
    public static class StatData {
        StatData(Map<String,Integer> names){
            myData = Collections.unmodifiableMap(names);
        }
        
        private StatData(String name){
            myData = Collections.singletonMap(name, 1);
        }
        
        public Map<String, Integer> getData(){
            return myData;
        }
        
        private StatData merge(StatData data){
            if ( data == null ){
                return this;
            }
            else {
                return new StatData(mergeAmounts(getData(), data.getData()));
            }
        }
        
        private final Map<String,Integer> myData; 
    }
    
}
