/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.mdr.persistence.memoryimpl;

import java.util.*;
import org.netbeans.mdr.persistence.*;

public class MultivalueLog {

    private static final byte OP_ADD = 0;
    private static final byte OP_REMOVE = 1;
    private static final byte OP_REPLACE = 2;
    private static final byte OP_ADD_ALL = 3;

    private static final int NULL_POSITION = -1;

    private ArrayList log = new ArrayList ();
    private MultivaluedIndexImpl index;
    private boolean rollingBack = false;

    // ..........................................................................    
    
    public MultivalueLog(MultivaluedIndexImpl index) {
        this.index = index;
    }
    
    // ..........................................................................
    
    /**
     * Clears the transaction log.
     */
    public void clear () {
        log.clear ();
    }

    /**
     * Logs add operation.
     */
    public void logAdd (Object key, Object value) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_REMOVE, key, value);
        log.add (0, rec);
    }

    /**
     * Logs add operation.
     */
    public void logAdd (Object key, Object value, int position) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_REMOVE, key, value, position);
        log.add (0, rec);
    }
    
    /**
     * Logs remove operation.
     */
    public void logRemove (Object key, Object value) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_ADD, key, value);
        log.add (0, rec);
    }

    /**
     * Logs remove operation.
     */
    public void logRemove (Object key, Object value, int position) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_ADD, key, value, position);
        log.add (0, rec);
    }
    
    /**
     * Logs remove operation.
     */
    public void logRemoveKey (Object key, Object value) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_ADD_ALL, key, value);
        log.add (0, rec);
    }
    
    /**
     * Logs replace operation.
     */
    public void logReplace (Object key, Object oldValue, int position) {
        if (rollingBack)
            return;
        Record rec = new Record (OP_REPLACE, key, oldValue, position);
        log.add (0, rec);
    }
    
    /**
     * Rolls back all changes performed over the index during the recorded transaction.
     */
    public void rollBack () throws StorageException {
        rollingBack = true;
        Iterator iter = log.iterator();
        while(iter.hasNext()) {            
            Record rec = (Record) iter.next();
            switch (rec.opCode) {
                case OP_ADD:
                    if (rec.position == NULL_POSITION)
                        index.add (rec.key, rec.value);
                    else
                        ((MultivaluedOrderedIndex) index).add (rec.key, rec.position, rec.value);
                    break;
                case OP_REMOVE:
                    if (rec.position == NULL_POSITION)
                        index.remove (rec.key, rec.value);
                    else
                        ((MultivaluedOrderedIndex) index).remove (rec.key, rec.position);
                    break;
                case OP_REPLACE:
                    ((MultivaluedOrderedIndex) index).replace (rec.key, rec.position, rec.value);
                    break;
                case OP_ADD_ALL:
                    index.setKey (rec.key, rec.value);
            }
        }
        rollingBack = false;
    }

    // Record ...................................................................

    private static class Record {
        
        byte opCode; // operation code
        Object key;
        Object value;
        int position;

        public Record (byte opCode, Object key, Object value, int position) {
            this.opCode = opCode;
            this.key = key;
            this.value = value;
            this.position = position;
        }
        
        public Record (byte opCode, Object key, Object value) {
            this.opCode = opCode;
            this.key = key;
            this.value = value;
            this.position = NULL_POSITION;
        }
        
    } // Record

}
