/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.mdr.storagemodel;

import java.util.*;

import org.netbeans.mdr.persistence.*;
import org.netbeans.mdr.util.DebugException;
import org.netbeans.mdr.util.Logger;

/**
 * Instances of this class are immutable, live collections representing
 * results of index access.
 *
 * @author Martin Matula
 */
public class IndexImmutSet implements Collection {

    /* --------------------------------------------------------------------- */
    /* -- Attributes (protected) ------------------------------------------- */
    /* --------------------------------------------------------------------- */
    
    protected final MdrStorage storage;
    protected final MultivaluedIndex index;
    protected final Object indexKey;
    
    /* --------------------------------------------------------------------- */
    /* -- Constructor (protected) ------------------------------------------ */
    /* --------------------------------------------------------------------- */
    
    /**
     * @param storage
     * @param index  the index where to look up the members of the current collection
     * @param indexKey the key under which to look up the members of the current collection
     */
    protected IndexImmutSet(MdrStorage storage, MultivaluedIndex index, Object indexKey) {
        this.storage = storage;
        this.index = index;
        this.indexKey = indexKey;
    }

    /* --------------------------------------------------------------------- */
    /* -- Methods for index access ----------------------------------------- */
    /* --------------------------------------------------------------------- */
    
    protected Collection getItems() {
        try {
            return index.getItems(indexKey);
        } catch (StorageException e) {
            throw (DebugException) Logger.getDefault().annotate(new DebugException(), e);
        }
    }
    
    protected Collection getObjects() {
        try {
            return storage.getObjectsFromIndex(index, indexKey);
        } catch (StorageException e) {
            throw (DebugException) Logger.getDefault().annotate(new DebugException(), e);
        }
    }
    
    /* --------------------------------------------------------------------- */
    /* -- implements java.util.Collection ---------------------------------- */
    /* --------------------------------------------------------------------- */
    
    public int size() {
        return getItems().size();
    }
    
    public boolean contains(Object obj) {
        return getItems().contains(obj);
    }
    
    public Iterator iterator() {
        return new IndexImmutIterator(getObjects().iterator());
    }
    
    public boolean isEmpty() {
        return getItems().isEmpty();
    }
    
    public boolean containsAll(Collection collection) {
        return getItems().containsAll(collection);
    }
    
    public Object[] toArray(Object[] obj) {
        return getObjects().toArray(obj);
    }
    
    public Object[] toArray() {
        return getObjects().toArray();
    }
    
   /**
    * operation not supported
    */
    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }        
    
   /**
    * operation not supported
    */
    public boolean add(Object obj) {
        throw new UnsupportedOperationException();
    }
    
   /**
    * operation not supported
    */
    public boolean removeAll(Collection collection) {
        throw new UnsupportedOperationException();
    }
    
   /**
    * operation not supported
    */
    public boolean addAll(Collection collection) {
        throw new UnsupportedOperationException();
    }
    
   /**
    * operation not supported
    */
    public boolean retainAll(Collection collection) {
        throw new UnsupportedOperationException();
    }
    
   /**
    * operation not supported
    */
    public void clear() {
        throw new UnsupportedOperationException();
    }
    
    /* --------------------------------------------------------------------- */
    /* -- IndexImmutIterator (inner class) --------------------------------- */
    /* --------------------------------------------------------------------- */
    
    /**
     * Wrapper for the index access iterator diabling modifying accesses.
     */
    protected class IndexImmutIterator implements Iterator {
        protected Iterator innerIterator;
        
        protected IndexImmutIterator(Iterator innerIterator) {
            this.innerIterator = innerIterator;
        }
        
        public boolean hasNext() {
            return innerIterator.hasNext();
        }
        
        public Object next() {
            return innerIterator.next();
        }
        
       /**
        * operation not supported
        */
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
