/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.cards.api;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;

/**
 *
 * @author Tim Boudreau
 */
public final class Card {

    private final Rank identity;
    private final Suits suit;
    private final Deck deck;

    public Card(Rank identity, Suits suit, Deck deck) {
        this.identity = identity;
        this.suit = suit;
        this.deck = deck;
        if (suit == null && identity != Rank.JOKER) {
            throw new IllegalArgumentException(identity + " " + suit + " " +
                    deck);
        }
    }

    public String getName() {
        if (isFaceDown()) {
            return NbBundle.getMessage(Card.class, "FACE_DOWN_CARD");
        }
        if (identity == Rank.JOKER) {
            return identity.toString();
        } else {
            return NbBundle.getMessage(Card.class, "FMT_CARD_NAME",
                    new Object[]{identity.toString(), suit.toString()});
        }
    }

    Rank identity() {
        return identity;
    }

    Suits suit() {
        return suit;
    }

    public Rank getRank() {
        if (isFaceDown()) {
            throw new IllegalStateException("Card is face down");
        }
        return identity;
    }

    public Suits getSuit() {
        if (isFaceDown()) {
            throw new IllegalStateException("Card is face down");
        }
        return suit;
    }

    public Deck getDeck() {
        return deck;
    }

    public boolean isFaceDown() {
        return pile == null ? false : !pile.isCardVisible(this);
    }

    Pile pile() {
        return pile;
    }

    public BufferedImage getImage(Dimension size) {
        BufferedImage img = getImage();
        if (size.width == img.getWidth() && size.height == img.getHeight()) {
            return img;
        } else {
            BufferedImage result = new BufferedImage(size.width, size.height,
                    BufferedImage.TYPE_INT_ARGB);
            Graphics2D g2d = result.createGraphics();
            g2d.setRenderingHint(RenderingHints.KEY_RENDERING,
                    RenderingHints.VALUE_RENDER_QUALITY);
            double sx = ((double) size.width) / ((double) img.getWidth());
            double sy = ((double) size.height) / ((double) img.getHeight());
            g2d.drawRenderedImage(img, AffineTransform.getScaleInstance(sx, sy));
            g2d.dispose();
            return result;
        }
    }

    private BufferedImage getImage() {
        return isFaceDown() ? getFaceDownImage() : getFaceUpImage();
    }

    private BufferedImage getFaceDownImage() {
        return deck.getFaceDownCardImage();
    }

    private BufferedImage getFaceUpImage() {
        return getImage(suit, identity);
    }
    Pile pile;

    void setPile(Pile pile) {
        if (pile != this.pile) {
            if (this.pile != null) {
                this.pile.doRemove(this);
            }
            this.pile = pile;
            if (pile != null) {
                pile.add(this);
            }
        }
    }

    public Pile getPile() {
        return pile;
    }

    @Override
    public boolean equals(Object o) {
        boolean result = o instanceof Card;
        if (result) {
            Card c = (Card) o;
            result = (c.suit() == null) == (suit() == null);
            if (result && c.suit() != null) {
                result = c.suit().equals(suit());
            }
            if (result) {
                result = c.identity().equals(identity()) &&
                        c.getDeck().equals(deck);
            }
        }
        return result;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 79 * hash + (this.identity != null ? this.identity.hashCode() : 0);
        hash = 79 * hash + (this.suit != null ? this.suit.hashCode() : 0);
        hash = 79 * hash + (this.deck != null ? this.deck.hashCode() : 0);
        return hash;
    }

    @Override
    public String toString() {
        return getName();
    }

    private static BufferedImage getImage(Suits suit, Rank rank) {
        return (BufferedImage) ImageUtilities.loadImage(getImageFileName(suit, rank));
    }

    private static String getImageFileName(Suits suit, Rank rank) {
        //From http://www.eludication.org/playingcards.html, creative
        //commons license
        String result;
        if (suit == null) {
            result = rank.imageNameInfix() + "-150.png";
        } else {
            result = suit.imageNamePrefix() + '-' + rank.imageNameInfix() +
                    "-150.png";
        }
        result = "org/netbeans/examples/cards/resources/" + result;
        return result;
    }
}
