/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.modules.manifesterrorannotation;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.jar.Manifest;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.StyledDocument;
import org.openide.cookies.EditCookie;
import org.openide.cookies.EditorCookie;
import org.openide.cookies.OpenCookie;
import org.openide.cookies.SaveCookie;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.text.Annotation;
import org.openide.text.DataEditorSupport;
import org.openide.util.RequestProcessor;

/**
 *
 * @author Jaroslav Tulach
 */
public class ManifestEditor extends DataEditorSupport
        implements OpenCookie, EditorCookie, EditCookie, DocumentListener, Runnable {

    private static Logger LOG = Logger.getLogger(ManifestEditor.class.getName());
    private static Pattern GOOD_LINE = Pattern.compile(" *|[\\p{Alnum}-._]+: *[\\p{Alnum}-/._]+");  // NOI18N
    private static RequestProcessor RP = new RequestProcessor("findError"); // NOI18N
    final RequestProcessor.Task findError;

    private ManifestEditor(ManifestDataObject obj) {
        super(obj, new MyEnv(obj));
        findError = RP.create(this, true);
        findError.setPriority(Thread.MIN_PRIORITY);
    }

    public static ManifestEditor create(ManifestDataObject obj) {
        return new ManifestEditor(obj);
    }

    protected boolean notifyModified() {
        boolean retValue;
        retValue = super.notifyModified();
        if (retValue) {
            ManifestDataObject obj = (ManifestDataObject) getDataObject();
            obj.ic.add(env);
        }

        getDocument().addDocumentListener(this);
        findError.schedule(0);

        return retValue;
    }

    protected void notifyUnmodified() {
        getDocument().removeDocumentListener(this);

        super.notifyUnmodified();

        ManifestDataObject obj = (ManifestDataObject) getDataObject();
        obj.ic.remove(env);
    }

    final void checkErrors() {
        StyledDocument d = getDocument();
        if (d == null) {
            // if parsing succeeds than there is no error
            ManifestErrorAnnotation.DEFAULT.detach();
            return;
        }
        String txt;
        try {
            txt = d.getText(0, d.getLength());
        } catch (BadLocationException ex) {
            LOG.log(Level.WARNING, null, ex);
            return;

        }
        InputStream is = new ByteArrayInputStream(txt.getBytes());
        try {
            new Manifest(is);
            // if parsing succeeds than there is no error
            ManifestErrorAnnotation.DEFAULT.detach();
            return;
        } catch (IOException ex) {
            // failure we need to place somewhere the annotation
        }

        String[] lines = txt.split("\n");
        for (int i = 0; i < lines.length; i++) {
            String line = lines[i];

            if (!GOOD_LINE.matcher(line).matches()) {
                ManifestErrorAnnotation.DEFAULT.attach(getLineSet().getCurrent(i));
                return;
            }
        }
    }

    public void insertUpdate(DocumentEvent e) {
        findError.schedule(500);
    }

    public void removeUpdate(DocumentEvent e) {
        findError.schedule(500);
    }

    public void changedUpdate(DocumentEvent e) {
        findError.schedule(500);
    }

    public void run() {
        checkErrors();
    }

    public StyledDocument openDocument() throws IOException {
        StyledDocument doc = super.openDocument();
        checkErrors();
        return doc;
    }

    private static final class ManifestErrorAnnotation extends Annotation {

        static final ManifestErrorAnnotation DEFAULT = new ManifestErrorAnnotation();

        public String getAnnotationType() {
            return "org-netbeans-modules-manifesterrorannotation";
        }

        public String getShortDescription() {
            return "manifest-error";
        }
    }

    private static final class MyEnv extends DataEditorSupport.Env
            implements SaveCookie {

        public MyEnv(ManifestDataObject obj) {
            super(obj);
        }

        public void save() throws IOException {
            ManifestEditor ed = (ManifestEditor) this.findCloneableOpenSupport();
            ed.saveDocument();
        }

        protected FileObject getFile() {
            return super.getDataObject().getPrimaryFile();
        }

        protected FileLock takeLock() throws IOException {
            return ((ManifestDataObject) super.getDataObject()).getPrimaryEntry().takeLock();
        }
    }
}
