/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.examples.modules.minicomposer;

import java.io.*;
import java.util.*;
import javax.sound.sampled.*;

/** An implementation of JavaSound's TargetDataLine,
 * generating the waveform from a Score.
 */
public class LineInFromScore implements TargetDataLine {
    /** An audio input stream based on the line in.
     * Computes the frame length based on some sampling rate.
     */
    public static class ScoreAudioInputStream extends AudioInputStream {
        public ScoreAudioInputStream(LineInFromScore line, float sampleRate) {
            super(line);
            Score s = line.s;
            frameLength = 0;
            for (int i = 0; i < s.getSize(); i++)
                frameLength += (int)(Score.DURATION_SECONDS[s.getNote(i).getDuration()] * sampleRate);
        }
    }
    private AudioFormat format;
    private float sampleRate;
    private Set listeners;
    private int frame;
    private final Score s;
    private int note;
    private int framesInNote;
    private int posInNote;
    private float frameToRadian;
    private boolean atEnd;
    private boolean isRest;
    public LineInFromScore(Score s) {
        this.s = s;
        listeners = new HashSet();
        frame = 0;
        note = -1;
        atEnd = false;
        sampleRate = ComposerSettings.getSampleRate();
        format = new AudioFormat(sampleRate, 8, 1, true, /* irrelevant */ false);
        update();
    }
    public int read(byte[] b, int off, int len) {
        for (int i = 0; i < len; i++) {
            if (atEnd) return i;
            frame++;
            b[off + i] = (isRest || posInNote * 10 > framesInNote * 9) ? (byte)0 :
                (byte)(127.0f * Math.sin(posInNote * frameToRadian));
            if (++posInNote == framesInNote) update();
        }
        return len;
    }
    private void update() {
        note++;
        if (note >= s.getSize()) {
            atEnd = true;
            return;
        }
        Score.Note n = s.getNote(note);
        framesInNote = (int)(Score.DURATION_SECONDS[n.getDuration()] * sampleRate);
        posInNote = 0;
        if (n.getTone() == 0) {
            // Rest.
            isRest = true;
        } else {
            isRest = false;
            float hertz = (float)(Score.MIDDLE_C_HERTZ *
                Math.pow(Score.HALF_STEP, n.getTone() - Score.WHERE_IS_C_TONE) *
                Math.pow(2.0, n.getOctave() - Score.WHERE_IS_MIDDLE_OCTAVE));
            frameToRadian = 2.0f * (float)Math.PI / sampleRate * hertz;
        }
    }
    public int available() {
        if (atEnd)
            return 0;
        else
            return framesInNote - posInNote;
    }
    public synchronized void addLineListener(LineListener listener) {
        listeners.add(listener);
    }
    public synchronized void removeLineListener(LineListener listener) {
        listeners.remove(listener);
    }
    protected synchronized void fireEvent(LineEvent.Type type) {
        if (listeners.isEmpty()) return;
        LineEvent ev = new LineEvent(this, type, frame);
        Iterator it = listeners.iterator();
        while (it.hasNext()) {
            LineListener listener = (LineListener)it.next();
            listener.update(ev);
        }
    }
    public boolean isOpen() {
        return true;
    }
    public void open() throws LineUnavailableException {
        fireEvent(LineEvent.Type.OPEN);
    }
    public void open(AudioFormat format) throws LineUnavailableException {
        if (!format.equals(this.format)) throw new LineUnavailableException();
        open();
    }
    public void open(AudioFormat format, int bufferSize) throws LineUnavailableException {
        open(format);
    }
    public Control getControl(Control.Type control) {
        throw new IllegalArgumentException();
    }
    public void flush() {
    }
    public void close() {
        fireEvent(LineEvent.Type.CLOSE);
    }
    public boolean isControlSupported(Control.Type control) {
        return false;
    }
    public Line.Info getLineInfo() {
        return new DataLine.Info(LineInFromScore.class, format, 1024);
    }
    public int getFramePosition() {
        return frame;
    }
    public long getLongFramePosition() { // JDK 1.5
        return frame;
    }
    public boolean isActive() {
        return true;
    }
    public void drain() {
    }
    public boolean isRunning() {
        return true;
    }
    public void stop() {
    }
    public float getLevel() {
        return 1.0f;
    }
    public Control[] getControls() {
        return new Control[] {};
    }
    public int getBufferSize() {
        return 1024;
    }
    public void start() {
    }
    public AudioFormat getFormat() {
        return format;
    }
    public long getMicrosecondPosition() {
        return (long)(1000000.0f * frame / sampleRate);
    }
}
