/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.modules.povproject;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.prefs.Preferences;
import javax.swing.event.ChangeListener;
import org.netbeans.modules.examples.api.povray.MainFileProvider;
import org.netbeans.modules.examples.api.povray.RendererService;
import org.openide.DialogDescriptor;
import org.openide.DialogDisplayer;
import org.openide.ErrorManager;
import org.openide.explorer.ExplorerManager;
import org.openide.filesystems.FileAttributeEvent;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileRenameEvent;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.Repository;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.util.WeakListeners;

/**
 * Implementation of RendererService for POV-Ray projects.
 *
 * @author Tim Boudreau
 */
final class RendererServiceImpl extends RendererService implements FileChangeListener {
    private PovrayProject proj;

    public RendererServiceImpl(PovrayProject proj) {
        this.proj = proj;
    }

    PovrayProject getProject() {
        return proj;
    }

    public FileObject render() {
        MainFileProvider mfp = proj.getLookup().lookup(
                MainFileProvider.class);
        assert mfp != null;
        if (mfp.getMainFile() == null) {
            showChooseMainFileDlg (mfp);
        }
        if (mfp.getMainFile() != null) {
            return render (null);
        } else {
            return null;
        }
    }

    private void showChooseMainFileDlg (final MainFileProvider mfp) {
        final MainFileChooser chooser = new MainFileChooser (proj);
        String title = NbBundle.getMessage(RendererServiceImpl.class,
                "TTL_ChooseMainFile");

        //Create a simple dialog descriptor describing what kind of dialog
        //we want and its title and contents
        final DialogDescriptor desc = new DialogDescriptor (chooser, title);

        //The OK button should be disabled initially
        desc.setValid(false);

        //Create a property change listener.  It will listen on the selection
        //in our MainFileChooser, and enable the OK button if an appropriate
        //node is selected:
        PropertyChangeListener pcl = new PropertyChangeListener() {
            public void propertyChange (PropertyChangeEvent pce) {

                String propName = pce.getPropertyName();

                if (ExplorerManager.PROP_SELECTED_NODES.equals(propName)) {
                    Node[] n = (Node[]) pce.getNewValue();

                    boolean valid = n.length == 1;

                    if (valid) {
                        DataObject ob = 
                            n[0].getLookup().lookup(DataObject.class);

                        valid = ob != null;

                        if (valid) {
                            FileObject selectedFile = ob.getPrimaryFile();
                            String mimeType = selectedFile.getMIMEType();
                            valid = "text/x-povray".equals(mimeType);
                        }
                    }
                    desc.setValid(valid);
                }
            }
        };
        chooser.getExplorerManager().addPropertyChangeListener(pcl);

        Object dialogResult = DialogDisplayer.getDefault().notify(desc);

        if (DialogDescriptor.OK_OPTION.equals(dialogResult)) {
            Node[] n = chooser.getExplorerManager().getSelectedNodes();
            DataObject ob = n[0].getLookup().lookup (DataObject.class);
            FileObject selectedFile = ob.getPrimaryFile();
            mfp.setMainFile(selectedFile);
        }
    }

    public FileObject render(FileObject scene) {
        return render (scene, (Properties) null);
    }

    public FileObject render(FileObject scene, String propertiesName) {
        assert Arrays.asList(
                getAvailableRendererSettingsNames()).contains(propertiesName);
        setPreferredRendererSettingsName(propertiesName);
        return render (scene, getRendererSettings(propertiesName));
    }


    public FileObject render(FileObject scene, Properties renderSettings) {
        Povray pov = new Povray (this, scene, renderSettings);
        FileObject result;
        try {
            result = pov.render();
            if (!listeningToImagesFolder) {
                startListeningToImagesFolder();
            }
        } catch (IOException ioe) {
            ErrorManager.getDefault().notify(ErrorManager.USER, ioe);
            result = null;
        }
        return result;
    }

    public String[] getAvailableRendererSettingsNames() {
        FileObject settingsFolder = getRendererSettingsFolder();
        String[] result;
        if (settingsFolder != null) {
            //Use a DataFolder here, so our ordering attributes in the layer
            //file are applied, and our returned String array will be in the
            //order we want
            DataFolder fld = DataFolder.findFolder(settingsFolder);
            DataObject[] kids = fld.getChildren();
            result = new String[ kids.length ];
            for (int i = 0; i < kids.length; i++) {
                result[i] = kids[i].getPrimaryFile().getNameExt();
            }
        } else {
            result = new String[0];
        }
        return result;
    }

    public String getDisplayName(String settingsName) {
        FileObject file = fileFor (settingsName);
        String result;
        if (file != null) {
            DataObject dob;
            try {
                dob = DataObject.find(file);
                result = dob.getNodeDelegate().getDisplayName();
            } catch (DataObjectNotFoundException ex) {
                //Should never happen
                ErrorManager.getDefault().notify (ex);
                result = "[error]";
            }
        } else {
            result = settingsName;
        }
        return result;
    }

    public Properties getRendererSettings(String name) {
        Properties result = new Properties();
        FileObject settingsFile = fileFor (name);
        if (settingsFile != null) {
            try {
                result.load(new BufferedInputStream(settingsFile.getInputStream()));
            } catch (FileNotFoundException ex) {
                ErrorManager.getDefault().notify (ex);
            } catch (IOException ex) {
                ErrorManager.getDefault().notify (ex);
            }
        } else {
            ErrorManager.getDefault().notify (ErrorManager.INFORMATIONAL,
                    new NullPointerException("Requested non-existent settings " +
                    "file " + name));
        }
        return result;
    }

    public String getPreferredRendererSettingsName() {
        String result = getPreferences().get(KEY_PREFERRED_SETTINGS, null);
        if (result == null) {
            result = "640x480.properties";
        }
        return result;
    }

    private static final String KEY_PREFERRED_SETTINGS = "preferredSettings";
    private void setPreferredRendererSettingsName(String val) {
        getPreferences().put(KEY_PREFERRED_SETTINGS, val);
    }

    static Preferences getPreferences() {
        return Preferences.userNodeForPackage(RendererServiceImpl.class);
    }

    private FileObject getRendererSettingsFolder() {
        String folderName = "Povray/RendererSettings/";
        FileSystem systemFileSystem =
                Repository.getDefault().getDefaultFileSystem();

        FileObject result =
                systemFileSystem.getRoot().getFileObject(folderName);

        if (result == null && !logged) {
            //Corrupted userdir or something is very very wrong.
            //Log it and move on.
            ErrorManager.getDefault().notify (ErrorManager.INFORMATIONAL,
                    new IllegalStateException("Renderer settings dir missing!"));
            logged = true;
        }
        return result;
    }

    private static boolean logged = false;
    private FileObject fileFor (String settingsName) {
        FileObject settingsFolder = getRendererSettingsFolder();
        FileObject result;
        if (settingsFolder != null) { //should never be null
            result = settingsFolder.getFileObject(settingsName);
        } else {
            result = null;
        }
        return result;
    }

    //Keep a list of the paths we are currently listening to
    private Set<String> scenesListenedTo = new HashSet<String>();
    private boolean listeningToImagesFolder = false;
    protected void listenerAdded(FileObject scene, ChangeListener l) {
        synchronized (this) {
            if (scenesListenedTo.add (scene.getPath())) {
                if (scenesListenedTo.size() == 1 || !listeningToImagesFolder) {
                    //This is the first call, so we should start listening
                    //on the images folder
                    startListeningToImagesFolder();
                }
                listenTo (scene);
            }
        }
    }

    protected void noLongerListeningTo(FileObject scene) {
        synchronized (this) {
            scenesListenedTo.remove (scene.getPath());
        }
    }

    private void startListeningToImagesFolder() {
        FileObject imageFolder = proj.getImagesFolder(false);
        listeningToImagesFolder = imageFolder != null;
        if (listeningToImagesFolder) {
            listenTo (imageFolder);
        }
    }

    private void listenTo (FileObject file) {
        //Add ourselves as a weak listener to the file.  This way we can still
        //be garbage collected if the project is closed
        FileChangeListener stub = WeakListeners.create(
                FileChangeListener.class, this, file);

        file.addFileChangeListener(stub);
    }

    public void fileFolderCreated(FileEvent fileEvent) {
        //Do nothing
    }

    public void fileDataCreated(FileEvent fileEvent) {
        FileObject created = fileEvent.getFile();
        fireSceneChange(created);
    }

    public void fileChanged(FileEvent fileEvent) {
        FileObject changed = fileEvent.getFile();
        fireSceneChange(changed);
    }

    public void fileDeleted(FileEvent fileEvent) {
        FileObject deleted = fileEvent.getFile();
        fireSceneChange(deleted);
        if (deleted.isFolder() && "images".equals(deleted.getNameExt())) {
            //The images folder was deleted, reset our listening flags
            fireSceneChange(null);
            listeningToImagesFolder = false;
        }
    }

    public void fileRenamed(FileRenameEvent fileRenameEvent) {
        //do nothing
    }

    public void fileAttributeChanged(FileAttributeEvent fileAttributeEvent) {
        //do nothing
    }
}
