/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 * 
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 * 
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.qa.reports;

import java.io.*;
import java.util.*;

/**
 * Issue from Issuezilla. This class holds information about one issue
 * in issuezilla.
 *
 * @author Jan Stola
 */
public class Issue implements Serializable {
    /** Number of the issue. */
    public int number;
    /** Status of the issue. */
    public int status;
    /** Mask for unconfirmed issues. */
    public static final int STATUS_UNCONFIRMED=1;
    /** Mask for new issues. */
    public static final int STATUS_NEW=2;
    /** Mask for assigned issues. */
    public static final int STATUS_ASSIGNED=4;
    /** Mask for reopened issues. */
    public static final int STATUS_REOPENED=8;
    /** Mask for resolved issues. */
    public static final int STATUS_RESOLVED=16;
    /** Mask for verified issues. */
    public static final int STATUS_VERIFIED=32;
    /** Mask for closed issues. */
    public static final int STATUS_CLOSED=64;
    /** Mask for open issues. */
    public static final int STATUS_OPEN=STATUS_UNCONFIRMED|STATUS_NEW|STATUS_ASSIGNED|STATUS_REOPENED;
    /** Priority of the issue. */
    public int priority;
    /** Mask for issues with priority P1. */
    public static final int PRIORITY_P1=1;
    /** Mask for issues with priority P2. */
    public static final int PRIORITY_P2=2;
    /** Mask for issues with priority P3. */
    public static final int PRIORITY_P3=4;
    /** Mask for issues with priority P4. */
    public static final int PRIORITY_P4=8;
    /** Mask for issues with priority P5. */
    public static final int PRIORITY_P5=16;
    /** Resolution of the issue. */
    public int resolution;
    /** Mask for issues with resolution <code>FIXED</code>. */
    public static final int RESOLUTION_FIXED=1;
    /** Mask for issues with resolution <code>INVALID</code>. */
    public static final int RESOLUTION_INVALID=2;
    /** Mask for issues with resolution <code>WONTFIX</code>. */
    public static final int RESOLUTION_WONTFIX=4;
    /** Mask for issues with resolution <code>LATER</code>. */
    public static final int RESOLUTION_LATER=8;
    /** Mask for issues with resolution <code>REMIND</code>. */
    public static final int RESOLUTION_REMIND=16;
    /** Mask for issues with resolution <code>DUPLICATE</code>. */
    public static final int RESOLUTION_DUPLICATE=32;
    /** Mask for issues with resolution <code>WORKSFORME</code>. */
    public static final int RESOLUTION_WORKSFORME=64;
    /** Mask for issues with resolution <code>MOVED</code>. */
    public static final int RESOLUTION_MOVED=128;
    /** Module this issue belongs to. */
    public String product;
    /** Version of the IDE this issue appears in. */
    public String version;
    /** Platform this issue appears on. */
    public String platform;
    /** Who is assigned to this issue. */
    public String assignedTo;
    /** Component this issue belongs to. */
    public String component;
    /** Reporter of this issue. */
    public String reporter;
    /** Target milestone for this issue. */
    public String target;
    /** Type of this issue (e.g. defect, enhancement, etc.) */
    public int type;
    /** Mask for issues with type <code>DEFECT</code>. */
    public static final int TYPE_DEFECT=1;
    /** Mask for issues with type <code>ENHANCEMENT</code>. */
    public static final int TYPE_ENHANCEMENT=2;
    /** Mask for issues with type <code>FEATURE</code>. */
    public static final int TYPE_FEATURE=4;
    /** Mask for issues with type <code>TASK</code>. */
    public static final int TYPE_TASK=8;
    /** Mask for issues with type <code>PATCH</code>. */
    public static final int TYPE_PATCH=16;
    /** Creation time of the issue. */
    public Date creation;
    /** QA contact for this issue. */
    public String QAContact;
    /** Status whiteboard of this issue. */
    public String statusWhiteboard;
    /** Operating system this issue can be reproduced on. */
    public String os;
    /** Summary of this issue. */
    public String summary;
    /** Keywords of this issue. */
    public Set keywords;
    /** Comments of this issue. */
    public List comments;
    /** Activity of this issue. */
    private IssueActivity activity;
    
    /**
     * Returns activity of this issue.
     *
     * @return activity of this issue.
     */
    public IssueActivity getActivity() throws IOException {
        if (activity==null) {
            activity=Queries.getActivity(number);
        }
        return activity;
    }
    
    /**
     * Determines whether the priority of the issue has changed since
     * given <code>date</code>.
     *
     * @param date reference date
     * @return change of the priority since given <code>date</code>.
     */
    public int priorityChangeSince(Date date) throws IOException {
        return getActivity().priorityChangeSince(date);
    }

    /**
     * Determines whether this issue is new in the given set of <code>modules</code>
     * since given <code>date</code>.
     *
     * @param date reference date.
     * @param modules reference set of modules.
     * @return {@link STATUS_NEW} if this issue has been newly created,
     * {@link STATUS_REOPENED} if the issue has been reopened,
     * -1 if the issue has been moved into this set of modules from another module
     * or 0 otherwise.
     */    
    public int newSince(Date date, Set modules) throws IOException {
        /* Issue created after the given date. */
        if (creation.after(date)) return STATUS_NEW;
        boolean movedAway=false;
        boolean movedHere=false;
        boolean closed=false;
        boolean opened=false;
        Iterator iter=getActivity().iterator();
        while (iter.hasNext()) {
            Change ch=(Change)iter.next();
            if (date.before(ch.when)) {
                if (ch.what.equals(IssueQuery.CHANGED_FIELD_PRODUCT_S)) {
                    /* The bug was moved from this module. */
                    if (modules.contains(ch.oldValue)&&(!modules.contains(ch.newValue))) {
                        if (movedHere) {
                            movedHere=false;
                        } else {
                            movedAway=true;
                        }
                    }
                    /* The bug was moved into this module. */
                    if ((!modules.contains(ch.oldValue))&&modules.contains(ch.newValue)&&(!movedAway)) {
                        movedHere=true;
                    }
                }
                if (ch.what.equals(IssueQuery.CHANGED_FIELD_STATUS_S)) {
                    /* The bug was closed. */
                    if ((ch.newValue.equals(IssueQuery.STATUS_RESOLVED)||
                    ch.newValue.equals(IssueQuery.STATUS_VERIFIED)||
                    ch.newValue.equals(IssueQuery.STATUS_CLOSED))&&(
                    ch.oldValue.equals(IssueQuery.STATUS_NEW)||
                    ch.oldValue.equals(IssueQuery.STATUS_ASSIGNED)||
                    ch.oldValue.equals(IssueQuery.STATUS_REOPENED))) {
                        if (opened) {
                            opened=false;
                        } else {
                            closed=true;
                        }
                    }
                    /* The bug was reopened. */
                    if ((!closed)&&(ch.oldValue.equals(IssueQuery.STATUS_RESOLVED)||
                    ch.oldValue.equals(IssueQuery.STATUS_VERIFIED)||
                    ch.oldValue.equals(IssueQuery.STATUS_CLOSED))&&(
                    ch.newValue.equals(IssueQuery.STATUS_NEW)||
                    ch.newValue.equals(IssueQuery.STATUS_ASSIGNED)||
                    ch.newValue.equals(IssueQuery.STATUS_REOPENED))) {
                        opened=true;
                    }
                }
            }
        }
        if (opened) {
            return STATUS_REOPENED;
        } else if (movedHere) {
            return -1;
        } else return 0;        
    }
    
    /**
     * Determines whether this issue is new since given <code>date</code>.
     *
     * @param date reference date.
     * @return {@link STATUS_NEW} if this issue has been newly created,
     * {@link STATUS_REOPENED} if the issue has been reopened,
     * -1 if the issue has been moved into this module from another module
     * or 0 otherwise.
     */    
    public int newSince(Date date) throws IOException {
        Set modules=new HashSet(1);
        modules.add(product);
        return newSince(date, modules);
    }

    /**
     * Determines whether this issue has been resolved since given <code>date</code>.
     *
     * @param date reference date.     
     * @return {@link STATUS_RESOLVED} if this issue has been resolved,
     * -1 if the issue has been moved away from this module
     * or 0 otherwise.
     */    
    public int resolvedSince(Date date) throws IOException {
        Set modules=new HashSet(1);
        modules.add(product);
        return resolvedSince(date, modules);        
    }
    
    
    /**
     * Determines whether this issue has been resolved since given <code>date</code>
     * (from the point of view of the given set of <code>modules</code>).
     *
     * @param date reference date.
     * @param modules reference set of modules.
     * @return {@link STATUS_RESOLVED} if this issue has been resolved,
     * -1 if the issue has been moved away from this module
     * or 0 otherwise.
     */    
    public int resolvedSince(Date date, Set modules) throws IOException {
        boolean movedAway=false;
        boolean movedHere=false;
        boolean closed=false;
        boolean opened=false;
        Iterator iter=getActivity().iterator();
        while (iter.hasNext()) {
            Change ch=(Change)iter.next();
            if (date.before(ch.when)) {
                if (ch.what.equals(IssueQuery.CHANGED_FIELD_PRODUCT_S)) {
                    /* The bug was moved into this module. */
                    if ((!modules.contains(ch.oldValue))&&modules.contains(ch.newValue)) {
                        if (movedAway) {
                            movedAway=false;
                        } else {
                            movedHere=true;
                        }
                    }
                    /* The bug was moved from this module. */
                    if ((!modules.contains(ch.newValue))&&modules.contains(ch.oldValue)&&(!movedHere)) {
                        movedAway=true;
                    }
                }
                if (ch.what.equals(IssueQuery.CHANGED_FIELD_STATUS_S)) {
                    /* The bug was reopened. */
                    if ((ch.oldValue.equals(IssueQuery.STATUS_RESOLVED)||
                    ch.oldValue.equals(IssueQuery.STATUS_VERIFIED)||
                    ch.oldValue.equals(IssueQuery.STATUS_CLOSED))&&(
                    ch.newValue.equals(IssueQuery.STATUS_NEW)||
                    ch.newValue.equals(IssueQuery.STATUS_ASSIGNED)||
                    ch.newValue.equals(IssueQuery.STATUS_REOPENED))) {
                        if (closed) {
                            closed=false;
                        } else {
                            opened=true;
                        }
                    }                    
                    /* The bug was closed. */
                    if ((!opened)&&(ch.newValue.equals(IssueQuery.STATUS_RESOLVED)||
                    ch.newValue.equals(IssueQuery.STATUS_VERIFIED)||
                    ch.newValue.equals(IssueQuery.STATUS_CLOSED))&&(
                    ch.oldValue.equals(IssueQuery.STATUS_NEW)||
                    ch.oldValue.equals(IssueQuery.STATUS_ASSIGNED)||
                    ch.oldValue.equals(IssueQuery.STATUS_REOPENED))) {
                        closed=true;
                    }
                }
            }
        }
        if (closed) {
            return STATUS_RESOLVED;
        } else if (movedAway) {
            return -1;
        } else return 0;
    }
    
}
