/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.modules.quiz;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TimeZone;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.NoResultException;
import javax.persistence.Persistence;
import javax.persistence.Query;
import javax.servlet.http.HttpServletRequest;
import org.netbeans.modules.quiz.entities.NbUser;
import org.netbeans.modules.quiz.entities.Quiz;
import org.netbeans.modules.quiz.entities.Response;
import org.netbeans.modules.quiz.entities.ResponseItem;
import org.netbeans.modules.quiz.entities.Winners;

/**
 *
 * @author Jindrich Sedek
 */
public final class PersistenceUtils {

    public static final String MYSQL_PU = "MYSQLPU";
    public static final String TEST_PU = "TestPU";
    public static final String PERF_TU = "PerformancePU";
    private static final Logger LOG = Logger.getLogger(PersistenceUtils.class.getName());
    private static final int BUFF_SIZE = 1024 * 1024;
    private static String default_PU = MYSQL_PU;
    private static PersistenceUtils instance;
    private static Map<Class, Integer> maxIds = new HashMap<Class, Integer>(3);
    private String actualQuizContent;
    private Quiz actualQuiz = null;
    private EntityManagerFactory emf;

    public static void persist(Object entity) {
        EntityManager em = getInstance().createEntityManager();
        em.getTransaction().begin();
        em.persist(entity);
        em.getTransaction().commit();
        em.close();
    }

    public static void merge(Object entity) {
        EntityManager em = getInstance().createEntityManager();
        em.getTransaction().begin();
        em.merge(entity);
        em.getTransaction().commit();
        em.close();
    }

    private PersistenceUtils(String pu) {
        emf = Persistence.createEntityManagerFactory(pu);
    }

    public static PersistenceUtils getInstance() {
        if (instance == null) {
            instance = new PersistenceUtils(default_PU);
        }
        return instance;
    }

    public static void setDefaultPu(String pu) {
        default_PU = pu;
    }

    public EntityManager createEntityManager() {
        return emf.createEntityManager();
    }

    public static <T> List<T> getAll(Class<T> entityType) {
        EntityManager em = getInstance().createEntityManager();
        List<T> result = getInstance().getAll(em, entityType);
        em.close();
        return result;
    }
    
    public <T> List<T> getAll(EntityManager em, Class<T> entityType) {
        Query q = em.createQuery("Select e from " + entityType.getSimpleName() + " e ");
        List<T> result = (List<T>) q.getResultList();
        return result;
    }

    public static Integer getNextId(Class entity) {
        Integer maxId = null;
        synchronized (maxIds) {
            maxId = maxIds.get(entity);
            if (maxId == null) {
                EntityManager em = getInstance().createEntityManager();
                Query query = em.createQuery("SELECT MAX(e.id) FROM " + entity.getSimpleName() + " e");
                maxId = (Integer) query.getSingleResult();
                if (maxId == null) {
                    maxId = 0;
                }
                em.close();
            }
            ++maxId;
            maxIds.put(entity, maxId);
        }
        return maxId;
    }

    public static void close() {
        if (instance != null) {
            instance.emf.close();
        }
    }

    public String getQuizContent(Quiz quiz) {
        if ((actualQuiz != null) && (actualQuiz.equals(quiz))&&(actualQuizContent != null)) {
            return actualQuizContent;
        }
        String path = quiz.getContentPath();
        File contentFile = new File(path);
        char[] buffer = new char[BUFF_SIZE];
        try {
            BufferedReader reader = new BufferedReader(new FileReader(contentFile));
            int readResult = reader.read(buffer);
            if (readResult == -1) {
                LOG.severe("EMPTY STREAM");
            }
            if (readResult == BUFF_SIZE) {
                LOG.severe("BUFFER TOO SMALL");
            }
            reader.close();
            if (readResult > 0) {//non empty file
                String str = new String(buffer, 0, readResult - 1);
                if ((actualQuiz != null)&&(actualQuiz.equals(quiz))){
                    actualQuizContent = str;
                }
                return str;
            }
        } catch (IOException ioex) {
            LOG.log(Level.SEVERE, ioex.getMessage(), ioex);
        }
        return null;
    }

    public Quiz getActualQuiz() {
        if (actualQuiz == null) {
            reloadActualQuiz();
        }
        return actualQuiz;
    }

    public void reloadActualQuiz() {
        List<Quiz> list = getAll(Quiz.class);
        Collections.sort(list);
        Iterator<Quiz> it = list.iterator();
        Quiz actual = null;
        while ((it.hasNext()) && (actual == null)) {
            Quiz next = it.next();
            if ((compareIndianToNow(next.getStartdate()) <= 0) && (compareIndianToNow(next.getEnddate()) >= 0)) {
                actual = next;
            }
        }
        actualQuiz = actual;
        // invalidate actual quiz content
        actualQuizContent = null;
    }


    public static void addOpenQuizes(NbUser user, HttpServletRequest request) {
        List<Quiz> openQuizes = PersistenceUtils.findOpenQuizesForUser(user);
        if (openQuizes.size() > 0){
            List<Integer> openQuizesIds = new ArrayList<Integer>(openQuizes.size());
            for (Quiz elem : openQuizes) {
                openQuizesIds.add(elem.getId());
            }
            request.setAttribute("openQuizes", openQuizesIds);
        }
    }
    
    /**
     * 
     * @param date from local database to be conveted in an indian time
     * @return the value 0 if the time represented by the argument is equal to actual time
     * a value less than 0 if the time represented by the argument is before the actual time ;
     * and a value greater than 0 if the time represented by the argument is after the actual time. 
     */
    public static int compareIndianToNow(Date indiaDate){
        return compareIndianToLocal(indiaDate, new Date());
    }
    
    static int compareIndianToLocal(Date indiaDate, Date localDate){
        TimeZone zone = TimeZone.getTimeZone("Asia/Calcutta");
        GregorianCalendar indiaCalendar = new GregorianCalendar(zone); 
        indiaCalendar.clear();
        indiaCalendar.set(indiaDate.getYear()+1900, indiaDate.getMonth(), indiaDate.getDate(), indiaDate.getHours(), indiaDate.getMinutes());
        GregorianCalendar localCalendar = new GregorianCalendar();
        localCalendar.clear();
        localCalendar.set(localDate.getYear()+1900, localDate.getMonth(), localDate.getDate(), localDate.getHours(), localDate.getMinutes());
        return indiaCalendar.compareTo(localCalendar);
    }

    private static List<Quiz> findOpenQuizesForUser(NbUser user) {
        EntityManager em = getInstance().createEntityManager();
        Query query = em.createQuery("SELECT q FROM Quiz q WHERE q.closed IS NULL OR q.closed = 0");
        List<Quiz> openQuizes = query.getResultList();
        Iterator<Quiz> it = openQuizes.iterator();
        List<Quiz> previousQuizes = new ArrayList<Quiz>(openQuizes.size());
        while (it.hasNext()){
            Quiz quiz = it.next();
            if (compareIndianToNow(quiz.getStartdate()) < 0){
                previousQuizes.add(quiz);
            }
        }
        query = em.createQuery("SELECT r.quiz FROM Response r WHERE r.nbuser = :nbuser");
        query.setParameter("nbuser", user);
        List<Quiz> filledQuizes = query.getResultList();
        previousQuizes.removeAll(filledQuizes);
        em.close();
        return previousQuizes;
    }
    
    
    public static Collection<ResponseItem> findResponsesByUserEmailAndQuiz(String email, Quiz quiz) {
        EntityManager em = getInstance().createEntityManager();
        Query q = em.createNamedQuery("NbUser.findByEmail");
        q.setParameter("email", email);
        NbUser user = (NbUser) q.getSingleResult();
        q = em.createNamedQuery("Response.findByUserSubmit");
        q.setParameter("nbuser", user);
        q.setParameter("quiz", quiz);
        try {
            Response response = (Response) q.getSingleResult();
            Collection<ResponseItem> responses = response.getResponseItemCollection();
            return responses;
        } catch (NoResultException ex) {
            LOG.fine("No responses found for e-mail/quiz" + email + "/" + quiz.getId());
        } finally {
            em.close();
        }
        return null;
    }

    
    public static List<ResponseItem> findResponsesByQuiz(Quiz quiz) {
        EntityManager em = getInstance().createEntityManager();
        Query q = em.createNamedQuery("ResponseItem.findByQuiz");
        q.setParameter("quiz", quiz);
        List<ResponseItem> items = q.getResultList();
        em.close();
        return items;
    }
    
    
    public void updateQuizWinners() {
        List<Quiz> quizzes = PersistenceUtils.getAll(Quiz.class);
        for (Quiz quiz : quizzes) {
            // if quiz closed and the winner is not assigned yet
            if (compareIndianToNow(quiz.getEnddate()) < 0 && quiz.getWinners() == null) {
                List<NbUser> winners = quiz.getWinnerCandidatesList();
                // select the winner by lot and persist
                if (winners.size() > 0) {
                    int winnerIdx = new Random().nextInt(winners.size());
                    EntityManager em = createEntityManager();
                    em.getTransaction().begin();
                    Winners winner = new Winners(quiz, winners.get(winnerIdx));
                    quiz.setWinners(winner);
                    em.persist(winner);
                    em.merge(quiz);
                    em.getTransaction().commit();
                    LOG.info("WINNER of quizID: "+quiz.getId()+" is: "+winner.getNbuser().getEmail());
                    em.close();
                } else {
                    LOG.warning("No winners for Quiz ID: " + quiz.getId()+"!!!");
                }
            }
        }
    }
}
