/*
                    Sun Public License Notice
 
The contents of this file are subject to the Sun Public License Version
1.0 (the "License"). You may not use this file except in compliance with
the License. A copy of the License is available at http://www.sun.com/
 
The Original Code is the Regsup module.
The Initial Developer of the Original Code is Petr Slegr.
Portions created by Petr Slegr are Copyright (C) 2002.
All Rights Reserved.
 
Contributor(s): Petr Slegr.
 */

package org.netbeans.modules.regsup.auted;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;

/** Class specifying graphic look and properties of every VisualAutomatonElement
 * class.
 * <p>
 * You can override only base paint() method which draws element in normal
 * circumstances and all paintXXX() (e.g. paintSelected()) methods will reflect
 * this. In this case you have to assign <code>shape</code> field because it is
 * used by paintXXX() methods (e.g. to colorize element by some color). Also, you
 * should not call fill(Color) in your paint() because filling by some color do
 * paintXXX() methods.
 * <p>
 * However, second way is to override any of painting methods to change look of element
 * dependently only on your mind.
 * <p>
 *
 * @author  Petr Slegr
 * @version 15.9.01/2
 */
public abstract class ElementGeometry {
    
    public static final int SELECT_SQUARE_SIZE = 4;
    
    // !!! shouldn't be these field static?
    protected Color activeColor = Color.green;
    
    protected Color errorColor = Color.red;
    
    protected Color warningColor = Color.yellow;
    
    protected Color breakpointColor = new Color (155, 246, 245);
    
    /**
     */
    protected Shape shape;
    
    /** Not useful now.
     */
    protected ElementGeometry () {
    }
    
    /** Paints element in normal circumstances (i.e. not selected, not errorneous,
     * not active during debugging and so on).
     */
    public abstract void paint (Graphics2D g);
    
    
    /** Paints element that is treat as active by debugger.
     * By default fills the shape by activeColor (e.g. green) and then calls paint().
     */
    public void paintActive (Graphics2D g) {
        Color oldColor = g.getColor ();
        g.setColor (activeColor);
        g.fill (shape);
        g.setColor (oldColor);
        paint (g);
    }
    
    /** Test whether specified Point is within this geometry. To be used for example
     * to display some info when mouse entered element geometry.
     * By default for edges returns false due to their nature, otherwise returns
     * shape.contains (Point).
     */
    public abstract boolean isOn (Point position);
    
    /** Test whether specified Point is within this geometry. To be used for selecting
     * of elements by mouse.
     * By default returns shape.contains (Point). For edges to be overriden.
     */
    public abstract boolean isOnForSelection (Point position);
    
    /** Utility method
     */
    protected static Rectangle2D prepareSelectSquare (double x, double y) {
        //System.out.println("prepareSelectSquare: x: " + x + " y: " + y);
        Rectangle2D res = new Rectangle2D.Double (x, y, SELECT_SQUARE_SIZE, SELECT_SQUARE_SIZE);
        return res;
    }
    
    /** Utility method
     */
    protected void paintSelectSquare (Graphics2D g, double x, double y) {
        Color old = g.getColor ();
        g.setColor (Color.blue); // !! const
        g.fill (prepareSelectSquare (x, y));
        g.setColor (old);
    }
    
    /** Utility method
     */
    // !!! why is this method here? it could be in EdgeGeometry
    protected static GeneralPath prepareArrow (double alfa, double pointX, double pointY) {
        double cosAlfa = Math.cos (alfa);
        double sinAlfa = Math.sin (alfa);
        double ARROW_A = VisualAutomatonEdge.EdgeGeometry.ARROW_A;
        double ARROW_B = VisualAutomatonEdge.EdgeGeometry.ARROW_B;
        double vpx = ARROW_A * cosAlfa;
        double vpy = ARROW_A * sinAlfa;
        // change back to Java system
        // !!! proc ??? jinak nez v setArrow - sjednotit a v setArrow pouzivat tuto metodu
        PointPlus vRelPoint = new PointPlus (- vpx, + vpy); // !!!
        //System.out.println ("vRelPoint: " + vRelPoint);
        Point vPoint = vRelPoint.add (pointX, pointY);
        //System.out.println ("vPoint: " + vPoint);
        double x = ARROW_B * sinAlfa;
        double y = ARROW_B * cosAlfa; // ! note: x ~ sin, y ~ cos
        PointPlus a1RelPoint = new PointPlus (x, y);
        Point a1Point = a1RelPoint.add (vPoint);
        x = -x;
        y = -y;
        PointPlus a2RelPoint = new PointPlus (x, y);
        Point a2Point = a2RelPoint.add (vPoint);
        GeneralPath arrow;
        arrow = new GeneralPath ();
        arrow.moveTo ((float) pointX, (float) pointY );
        arrow.lineTo ((float) a1Point.getX (), (float) a1Point.getY ());
        arrow.lineTo ((float) a2Point.getX (), (float) a2Point.getY ());
        arrow.closePath ();
        return arrow;
    }
    
    /** Paints element that is treat as selected by editor.
     * By default calls paint() and after that draws four blue square around the element.
     */
    void paintSelected (Graphics2D g) {
        paint (g);
        RectangularShape boundShape = shape.getBounds2D ();
        double x1 = boundShape.getX ();
        double x2 = x1 + boundShape.getWidth ();
        double y1 = boundShape.getY ();
        double y2 = y1 + boundShape.getHeight ();
        Color old = g.getColor ();
        g.setColor (Color.blue); // !! const
        g.fill (prepareSelectSquare (x1 - SELECT_SQUARE_SIZE, y1 - SELECT_SQUARE_SIZE));
        g.fill (prepareSelectSquare (x2, y1 - SELECT_SQUARE_SIZE));
        g.fill (prepareSelectSquare (x1 - SELECT_SQUARE_SIZE, y2));
        g.fill (prepareSelectSquare (x2, y2));
        g.setColor (old);
    }
    
    /** Paints element that is treat as errorneous.
     * By default fills the shape by errorColor (e.g. red) and then calls paint().
     */
    public void paintError (Graphics2D g) {
        Color old = g.getColor ();
        g.setColor (errorColor);
        g.fill (shape);
        g.setColor (old);
        paint (g);
    }

    public void paintBreakpoint (Graphics2D g) {
        Color old = g.getColor ();
        g.setColor (breakpointColor);
        g.fill (shape);
        g.setColor (old);
        paint (g);
    }
    
    /** This method could be implemented in NodeGeometry, ContainerGeometry (that
     * implement ConnectableElementGeometry), but its impl would be the same and that's why
     * it is here
     */
    public Point getCenter () {
        PointPlus pp = new PointPlus (shape.getBounds2D ().getCenterX (),
        shape.getBounds2D ().getCenterY ());
        return pp;
    }
    
    /** For debug
     */
    static void paintSnapPoints (Point[] snaps, Graphics2D g) {
        for (int i = 0; i < snaps.length; i++) {
            Point snap = snaps[i];
            Ellipse2D circle = new Ellipse2D.Double (snap.getX (), snap.getY (), 2, 2);
            g.draw (circle);
        }
    }
}
